# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import json

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward.forth import ForthMachine32


def test_skip_ws():
    vm = ForthMachine32("input x  x skipws")
    vm.run({"x": np.array([ord(x) for x in "     hello"], np.uint8)})
    assert vm.input_position("x") == 5


def test_textint():
    vm = ForthMachine32(
        """input x output y float64
           x skipws 2 x #textint-> stack
           x skipws x textint-> stack
           x skipws 2 x #textint-> y"""
    )
    vm.run(
        {"x": np.array([ord(x) for x in "     12345 -123      3210  -42 0"], np.uint8)}
    )
    assert vm.stack == [12345, -123, 3210]
    assert np.asarray(vm["y"]).tolist() == [-42.0, 0.0]


def test_textfloat():
    vm = ForthMachine32(
        """input x output y float64
           x skipws 5 x #textfloat-> y
           x skipws x textfloat-> y"""
    )
    vm.run(
        {
            "x": np.array(
                [ord(x) for x in "     42 -42 42.123 -42.123e1 42.123e+2 -42.123e-2"],
                np.uint8,
            )
        }
    )
    assert np.asarray(vm["y"]).tolist() == pytest.approx(
        [42, -42, 42.123, -42.123e1, 42.123e2, -42.123e-2]
    )


def test_quotedstr():
    vm = ForthMachine32(
        """input x output y uint8
           x skipws 5 x #quotedstr-> y"""
    )
    vm.run(
        {
            "x": np.array(
                [
                    ord(x)
                    for x in r'     "one" "   two"   "three   " "fo\\u\"r"     "f\niv\u0045"'
                ],
                np.uint8,
            )
        }
    )
    assert (
        ak._v2._util.tobytes(np.asarray(vm["y"])) == b'one   twothree   fo\\u"rf\nivE'
    )
    assert vm.stack == [3, 6, 8, 6, 5]


def test_unicode():
    vm = ForthMachine32(
        """input x output y uint8
           x quotedstr-> y"""
    )
    x = np.array(
        [ord(x) for x in r'"\u0000 \u007f \u0080 \u07ff \u0800 \ud7ff \ue000 \uffff"'],
        np.uint8,
    )
    expecting = json.loads(ak._v2._util.tobytes(x)).encode("utf-8")
    vm.run({"x": x})
    assert ak._v2._util.tobytes(np.asarray(vm["y"])) == expecting
    assert vm.stack == [25] == [len(expecting)]


def test_peek():
    vm = ForthMachine32(
        """input x output y float64
           5 0 do x skipws 0 x peek 1 x peek x textint-> y loop"""
    )
    vm.run(
        {"x": np.array([ord(x) for x in "     12345 -123      3210  -42 98"], np.uint8)}
    )
    assert vm.stack == [
        ord("1"),
        ord("2"),
        ord("-"),
        ord("1"),
        ord("3"),
        ord("2"),
        ord("-"),
        ord("4"),
        ord("9"),
        ord("8"),
    ]
    assert np.asarray(vm["y"]).tolist() == [12345, -123, 3210, -42, 98]


def test_enum():
    vm = ForthMachine32(
        r'input x   3 0 do x skipws x enum s" zero" s" one" s" two" s" three" loop'
    )
    vm.run({"x": np.array([ord(x) for x in "one three two"], np.uint8)})
    assert vm.stack == [1, 3, 2]


def test_general_case():
    vm = ForthMachine32(
        r"case 0 of 1000 endof 1 of 1001 endof 1 1 + of 1002 endof 9999 swap endcase 10 20 30"
    )

    vm.begin()
    vm.stack_push(0)
    vm.resume()
    assert vm.stack == [1000, 10, 20, 30]

    vm.begin()
    vm.stack_push(1)
    vm.resume()
    assert vm.stack == [1001, 10, 20, 30]

    vm.begin()
    vm.stack_push(2)
    vm.resume()
    assert vm.stack == [1002, 10, 20, 30]

    vm.begin()
    vm.stack_push(3)
    vm.resume()
    assert vm.stack == [9999, 10, 20, 30]

    vm.begin()
    vm.stack_push(-1)
    vm.resume()
    assert vm.stack == [9999, 10, 20, 30]

    vm = ForthMachine32(
        r"case 0 of 1000 endof 1 of 1001 endof 1 1 + of 1002 endof endcase 10 20 30"
    )

    vm.begin()
    vm.stack_push(2)
    vm.resume()
    assert vm.stack == [1002, 10, 20, 30]

    vm.begin()
    vm.stack_push(3)
    vm.resume()
    assert vm.stack == [10, 20, 30]

    vm.begin()
    vm.stack_push(-1)
    vm.resume()
    assert vm.stack == [10, 20, 30]


def test_specialized_case():
    vm = ForthMachine32(
        r"case 0 of 1000 endof 1 of 1001 endof 2 of 1002 endof 9999 swap endcase 10 20 30"
    )

    vm.begin()
    vm.stack_push(0)
    vm.resume()
    assert vm.stack == [1000, 10, 20, 30]

    vm.begin()
    vm.stack_push(1)
    vm.resume()
    assert vm.stack == [1001, 10, 20, 30]

    vm.begin()
    vm.stack_push(2)
    vm.resume()
    assert vm.stack == [1002, 10, 20, 30]

    vm.begin()
    vm.stack_push(3)
    vm.resume()
    assert vm.stack == [9999, 10, 20, 30]

    vm.begin()
    vm.stack_push(-1)
    vm.resume()
    assert vm.stack == [9999, 10, 20, 30]

    vm = ForthMachine32(
        r"case 0 of 1000 endof 1 of 1001 endof 2 of 1002 endof endcase 10 20 30"
    )

    vm.begin()
    vm.stack_push(2)
    vm.resume()
    assert vm.stack == [1002, 10, 20, 30]

    vm.begin()
    vm.stack_push(3)
    vm.resume()
    assert vm.stack == [10, 20, 30]

    vm.begin()
    vm.stack_push(-1)
    vm.resume()
    assert vm.stack == [10, 20, 30]

    vm = ForthMachine32(r"case 9999 swap endcase 10 20 30")

    vm.begin()
    vm.stack_push(0)
    vm.resume()
    assert vm.stack == [9999, 10, 20, 30]

    vm = ForthMachine32(r"case endcase 10 20 30")

    vm.begin()
    vm.stack_push(0)
    vm.resume()
    assert vm.stack == [10, 20, 30]

    vm = ForthMachine32(
        r"case 0 of 1000 endof 1 of 1001 endof 2 of endof 9999 swap endcase 10 20 30"
    )

    vm.begin()
    vm.stack_push(1)
    vm.resume()
    assert vm.stack == [1001, 10, 20, 30]

    vm.begin()
    vm.stack_push(2)
    vm.resume()
    assert vm.stack == [10, 20, 30]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_numbers_to_type():
    one, two = (
        ak._v2.highlevel.Array([1, 2, 3]).layout,
        ak._v2.highlevel.Array([4, 5]).layout,
    )

    assert np.asarray(one.numbers_to_type("bool")).dtype == np.dtype(np.bool_)
    assert np.asarray(one.numbers_to_type("int8")).dtype == np.dtype(np.int8)
    assert np.asarray(one.numbers_to_type("uint8")).dtype == np.dtype(np.uint8)
    assert np.asarray(one.numbers_to_type("int16")).dtype == np.dtype(np.int16)
    assert np.asarray(one.numbers_to_type("uint16")).dtype == np.dtype(np.uint16)
    assert np.asarray(one.numbers_to_type("int32")).dtype == np.dtype(np.int32)
    assert np.asarray(one.numbers_to_type("uint32")).dtype == np.dtype(np.uint32)
    assert np.asarray(one.numbers_to_type("int64")).dtype == np.dtype(np.int64)
    assert np.asarray(one.numbers_to_type("uint64")).dtype == np.dtype(np.uint64)
    assert np.asarray(one.numbers_to_type("float32")).dtype == np.dtype(np.float32)
    assert np.asarray(one.numbers_to_type("float64")).dtype == np.dtype(np.float64)
    assert np.asarray(one.numbers_to_type("complex64")).dtype == np.dtype(np.complex64)
    assert np.asarray(one.numbers_to_type("complex128")).dtype == np.dtype(
        np.complex128
    )
    assert np.asarray(one.numbers_to_type("datetime64")).dtype == np.dtype(
        np.datetime64
    )
    assert np.asarray(one.numbers_to_type("datetime64[Y]")).dtype == np.dtype(
        "datetime64[Y]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[M]")).dtype == np.dtype(
        "datetime64[M]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[W]")).dtype == np.dtype(
        "datetime64[W]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[D]")).dtype == np.dtype(
        "datetime64[D]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[h]")).dtype == np.dtype(
        "datetime64[h]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[m]")).dtype == np.dtype(
        "datetime64[m]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[s]")).dtype == np.dtype(
        "datetime64[s]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[ms]")).dtype == np.dtype(
        "datetime64[ms]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[us]")).dtype == np.dtype(
        "datetime64[us]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[ns]")).dtype == np.dtype(
        "datetime64[ns]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[ps]")).dtype == np.dtype(
        "datetime64[ps]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[fs]")).dtype == np.dtype(
        "datetime64[fs]"
    )
    assert np.asarray(one.numbers_to_type("datetime64[as]")).dtype == np.dtype(
        "datetime64[as]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64")).dtype == np.dtype(
        np.timedelta64
    )
    assert np.asarray(one.numbers_to_type("timedelta64[Y]")).dtype == np.dtype(
        "timedelta64[Y]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[M]")).dtype == np.dtype(
        "timedelta64[M]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[W]")).dtype == np.dtype(
        "timedelta64[W]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[D]")).dtype == np.dtype(
        "timedelta64[D]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[h]")).dtype == np.dtype(
        "timedelta64[h]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[m]")).dtype == np.dtype(
        "timedelta64[m]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[s]")).dtype == np.dtype(
        "timedelta64[s]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[ms]")).dtype == np.dtype(
        "timedelta64[ms]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[us]")).dtype == np.dtype(
        "timedelta64[us]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[ns]")).dtype == np.dtype(
        "timedelta64[ns]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[ps]")).dtype == np.dtype(
        "timedelta64[ps]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[fs]")).dtype == np.dtype(
        "timedelta64[fs]"
    )
    assert np.asarray(one.numbers_to_type("timedelta64[as]")).dtype == np.dtype(
        "timedelta64[as]"
    )

    assert np.asarray(two.numbers_to_type("bool")).dtype == np.dtype(np.bool_)
    assert np.asarray(two.numbers_to_type("int8")).dtype == np.dtype(np.int8)
    assert np.asarray(two.numbers_to_type("uint8")).dtype == np.dtype(np.uint8)
    assert np.asarray(two.numbers_to_type("int16")).dtype == np.dtype(np.int16)
    assert np.asarray(two.numbers_to_type("uint16")).dtype == np.dtype(np.uint16)
    assert np.asarray(two.numbers_to_type("int32")).dtype == np.dtype(np.int32)
    assert np.asarray(two.numbers_to_type("uint32")).dtype == np.dtype(np.uint32)
    assert np.asarray(two.numbers_to_type("int64")).dtype == np.dtype(np.int64)
    assert np.asarray(two.numbers_to_type("uint64")).dtype == np.dtype(np.uint64)
    assert np.asarray(two.numbers_to_type("float32")).dtype == np.dtype(np.float32)
    assert np.asarray(two.numbers_to_type("float64")).dtype == np.dtype(np.float64)
    assert np.asarray(two.numbers_to_type("complex64")).dtype == np.dtype(np.complex64)
    assert np.asarray(two.numbers_to_type("complex128")).dtype == np.dtype(
        np.complex128
    )
    assert np.asarray(two.numbers_to_type("datetime64")).dtype == np.dtype(
        np.datetime64
    )
    assert np.asarray(two.numbers_to_type("datetime64[Y]")).dtype == np.dtype(
        "datetime64[Y]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[M]")).dtype == np.dtype(
        "datetime64[M]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[W]")).dtype == np.dtype(
        "datetime64[W]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[D]")).dtype == np.dtype(
        "datetime64[D]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[h]")).dtype == np.dtype(
        "datetime64[h]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[m]")).dtype == np.dtype(
        "datetime64[m]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[s]")).dtype == np.dtype(
        "datetime64[s]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[ms]")).dtype == np.dtype(
        "datetime64[ms]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[us]")).dtype == np.dtype(
        "datetime64[us]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[ns]")).dtype == np.dtype(
        "datetime64[ns]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[ps]")).dtype == np.dtype(
        "datetime64[ps]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[fs]")).dtype == np.dtype(
        "datetime64[fs]"
    )
    assert np.asarray(two.numbers_to_type("datetime64[as]")).dtype == np.dtype(
        "datetime64[as]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64")).dtype == np.dtype(
        np.timedelta64
    )
    assert np.asarray(two.numbers_to_type("timedelta64[Y]")).dtype == np.dtype(
        "timedelta64[Y]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[M]")).dtype == np.dtype(
        "timedelta64[M]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[W]")).dtype == np.dtype(
        "timedelta64[W]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[D]")).dtype == np.dtype(
        "timedelta64[D]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[h]")).dtype == np.dtype(
        "timedelta64[h]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[m]")).dtype == np.dtype(
        "timedelta64[m]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[s]")).dtype == np.dtype(
        "timedelta64[s]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[ms]")).dtype == np.dtype(
        "timedelta64[ms]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[us]")).dtype == np.dtype(
        "timedelta64[us]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[ns]")).dtype == np.dtype(
        "timedelta64[ns]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[ps]")).dtype == np.dtype(
        "timedelta64[ps]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[fs]")).dtype == np.dtype(
        "timedelta64[fs]"
    )
    assert np.asarray(two.numbers_to_type("timedelta64[as]")).dtype == np.dtype(
        "timedelta64[as]"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numbers
from collections.abc import Iterable

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def to_cupy(array):
    """
    Converts `array` (many types supported) into a CuPy array, if possible.

    If the data are numerical and regular (nested lists have equal lengths
    in each dimension, as described by the #type), they can be losslessly
    converted to a CuPy array and this function returns without an error.

    Otherwise, the function raises an error.

    If `array` is a scalar, it is converted into a CuPy scalar.

    See also #ak.from_cupy and #ak.to_numpy.
    """
    cupy = ak.nplike.Cupy.instance()
    np = ak.nplike.NumpyMetadata.instance()

    if isinstance(array, (bool, numbers.Number)):
        return cupy.array(array)

    elif isinstance(array, cupy.ndarray):
        return array

    elif isinstance(array, np.ndarray):
        return cupy.asarray(array)

    elif isinstance(array, ak._v2.highlevel.Array):
        return to_cupy(array.layout)

    elif isinstance(array, ak._v2.highlevel.Record):
        raise ValueError("CuPy does not support record structures")

    elif isinstance(array, ak._v2.highlevel.ArrayBuilder):
        return to_cupy(array.snapshot().layout)

    elif isinstance(array, ak.layout.ArrayBuilder):
        return to_cupy(array.snapshot())

    elif (
        ak._v2.operations.describe.parameters(array).get("__array__") == "bytestring"
        or ak._v2.operations.describe.parameters(array).get("__array__") == "string"
    ):
        raise ValueError("CuPy does not support arrays of strings")

    elif isinstance(array, ak._v2.contents.EmptyArray):
        return cupy.array([])

    elif isinstance(array, ak._v2.contents.IndexedArray):
        return to_cupy(array.project())

    elif isinstance(array, ak._v2.contents.UnionArray):
        contents = [to_cupy(array.project(i)) for i in range(len(array.contents))]
        out = cupy.concatenate(contents)

        tags = cupy.asarray(array.tags)
        for tag, content in enumerate(contents):
            mask = tags == tag
            out[mask] = content
        return out

    elif isinstance(array, ak._v2.contents.UnmaskedArray):
        return to_cupy(array.content)

    elif isinstance(array, ak._v2.contents.IndexedOptionArray):
        content = to_cupy(array.project())

        shape = list(content.shape)
        shape[0] = len(array)
        mask0 = cupy.asarray(array.bytemask()).view(np.bool_)
        if mask0.any():
            raise ValueError("CuPy does not support masked arrays")
        else:
            return content

    elif isinstance(array, ak._v2.contents.RegularArray):
        out = to_cupy(array.content)
        head, tail = out.shape[0], out.shape[1:]
        shape = (head // array.size, array.size) + tail
        return out[: shape[0] * array.size].reshape(shape)

    elif isinstance(
        array, (ak._v2.contents.ListArray, ak._v2.contents.ListOffsetArray)
    ):
        return to_cupy(array.toRegularArray())

    elif isinstance(array, ak._v2.contents.recordarray.RecordArray):
        raise ValueError("CuPy does not support record structures")

    elif isinstance(array, ak._v2.contents.NumpyArray):
        return array.to_cupy()

    elif isinstance(array, ak._v2.contents.Content):
        raise AssertionError(f"unrecognized Content type: {type(array)}")

    elif isinstance(array, Iterable):
        return cupy.asarray(array)

    else:
        raise ValueError(f"cannot convert {array} into cp.ndarray")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def firsts(array, axis=1, highlevel=True, behavior=None):
    raise NotImplementedError


#     """
#     Args:
#         array: Data from which to select the first elements from nested lists.
#         axis (int): The dimension at which this operation is applied. The
#             outermost dimension is `0`, followed by `1`, etc., and negative
#             values count backward from the innermost: `-1` is the innermost
#             dimension, `-2` is the next level up, etc.
#         highlevel (bool): If True, return an #ak.Array; otherwise, return
#             a low-level #ak.layout.Content subclass.
#         behavior (None or dict): Custom #ak.behavior for the output array, if
#             high-level.

#     Selects the first element of each non-empty list and inserts None for each
#     empty list.

#     For example,

#         >>> array = ak.Array([[1.1], [2.2], [], [3.3], [], [], [4.4], [5.5]])
#         >>> print(ak.firsts(array))
#         [1.1, 2.2, None, 3.3, None, None, 4.4, 5.5]

#     See #ak.singletons to invert this function.
#     """
#     layout = ak._v2.operations.convert.to_layout(
#         array, allow_record=False, allow_other=False
#     )
#     if isinstance(layout, ak.partition.PartitionedArray):   # NO PARTITIONED ARRAY
#         posaxis = None
#         for x in layout.partitions:
#             if posaxis is None:
#                 posaxis = x.axis_wrap_if_negative(axis)
#             elif posaxis != x.axis_wrap_if_negative(axis):
#                 raise ValueError(
#                     "ak.firsts for partitions with different axis depths"
#
#                 )
#     else:
#         posaxis = layout.axis_wrap_if_negative(axis)

#     if posaxis == 0:
#         if len(layout) == 0:
#             out = None
#         else:
#             out = layout[0]
#     else:
#         if posaxis < 0:
#             raise NotImplementedError(
#                 "ak.firsts with ambiguous negative axis"
#
#             )
#         toslice = (slice(None, None, None),) * posaxis + (0,)
#         out = ak.mask(layout, ak.num(layout, axis=posaxis) > 0, highlevel=False)[
#             toslice
#         ]

#     return ak._v2._util.maybe_wrap_like(out, array, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_byte_masked_array():
    layout = ak._v2.contents.ByteMaskedArray(
        ak._v2.index.Index8(np.r_[0, 1, 1]),
        ak._v2.contents.IndexedOptionArray(
            ak._v2.index.Index64([0, 1, -1]), ak._v2.contents.NumpyArray(np.arange(3))
        ),
        valid_when=True,
    )
    assert layout.tolist() == [None, 1, None]

    simplified = layout.simplify_optiontype()
    assert simplified.tolist() == [None, 1, None]

    assert isinstance(simplified, ak._v2.contents.IndexedOptionArray)
    assert np.asarray(simplified.index).tolist() == [-1, 1, -1]


def test_bit_masked_array():
    layout = ak._v2.contents.BitMaskedArray(
        ak._v2.index.IndexU8(np.array([0 | 1 << 1 | 1 << 2], dtype=np.uint8)),
        ak._v2.contents.IndexedOptionArray(
            ak._v2.index.Index64([0, 1, -1]), ak._v2.contents.NumpyArray(np.arange(3))
        ),
        valid_when=True,
        length=3,
        lsb_order=True,
    )
    assert layout.tolist() == [None, 1, None]

    simplified = layout.simplify_optiontype()
    assert simplified.tolist() == [None, 1, None]

    assert isinstance(simplified, ak._v2.contents.IndexedOptionArray)
    assert np.asarray(simplified.index).tolist() == [-1, 1, -1]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def from_parquet(  # this will be going through Awkward-Dask
    source,
    columns=None,
    row_groups=None,
    use_threads=True,
    include_partition_columns=True,
    lazy=False,
    lazy_cache="new",
    lazy_cache_key=None,
    highlevel=True,
    behavior=None,
    **options
):
    raise NotImplementedError


#     """
#     Args:
#         source (str, Path, file-like object, pyarrow.NativeFile): Where to
#             get the Parquet file. If `source` is the name of a local directory
#             (str or Path), then it is interpreted as a partitioned Parquet dataset.
#         columns (None or list of str): If None, read all columns; otherwise,
#             read a specified set of columns.
#         row_groups (None, int, or list of int): If None, read all row groups;
#             otherwise, read a single or list of row groups.
#         use_threads (bool): Passed to the pyarrow.parquet.ParquetFile.read
#             functions; if True, do multithreaded reading.
#         include_partition_columns (bool): If True and `source` is a partitioned
#             Parquet dataset with subdirectory names defining partition names
#             and values, include those special columns in the output.
#         lazy (bool): If True, read columns in row groups on demand (as
#             #ak.layout.VirtualArray, possibly in #ak.partition.PartitionedArray
#             if the file has more than one row group); if False, read all
#             requested data immediately.
#         lazy_cache (None, "new", or MutableMapping): If lazy, pass this
#             cache to the VirtualArrays. If "new", a new dict (keep-forever cache)
#             is created. If None, no cache is used.
#         lazy_cache_key (None or str): If lazy, pass this cache_key to the
#             VirtualArrays. If None, a process-unique string is constructed.
#         highlevel (bool): If True, return an #ak.Array; otherwise, return
#             a low-level #ak.layout.Content subclass.
#         behavior (None or dict): Custom #ak.behavior for the output array, if
#             high-level.
#         options: All other options are passed to pyarrow.parquet.ParquetFile.

#     Reads a Parquet file into an Awkward Array (through pyarrow).

#         >>> ak.from_parquet("array1.parquet")
#         <Array [[1, 2, 3], [], ... [], [6, 7, 8, 9]] type='6 * var * ?int64'>

#     See also #ak.from_arrow, which is used as an intermediate step.
#     See also #ak.to_parquet.
#     """



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from collections.abc import Iterable

import awkward as ak
from awkward._v2.contents.content import Content

np = ak.nplike.NumpyMetadata.instance()


class Record:
    def __init__(self, array, at):
        if not isinstance(array, ak._v2.contents.recordarray.RecordArray):
            raise TypeError(f"Record 'array' must be a RecordArray, not {array!r}")
        if not ak._util.isint(at):
            raise TypeError(f"Record 'at' must be an integer, not {array!r}")
        if at < 0 or at >= array.length:
            raise ValueError(
                f"Record 'at' must be >= 0 and < len(array) == {array.length}, not {at}"
            )
        else:
            self._array = array
            self._at = at

    @property
    def array(self):
        return self._array

    @property
    def at(self):
        return self._at

    @property
    def fields(self):
        return self._array.fields

    @property
    def is_tuple(self):
        return self._array.is_tuple

    @property
    def as_tuple(self):
        return Record(self._array.as_tuple, self._at)

    @property
    def contents(self):
        out = []
        for field in self._array.fields:
            out.append(self._array[field][self._at])
        return out

    def content(self, index_or_field):
        return self._array.content(index_or_field)[self._at]

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<Record at="]
        out.append(repr(str(self._at)))
        out.append(">\n")
        out.append(self._array._repr(indent + "    ", "<array>", "</array>\n"))
        out.append(indent)
        out.append("</Record>")
        out.append(post)
        return "".join(out)

    def validityerror(self, path="layout.array"):
        return self._array.validityerror(path)

    @property
    def parameters(self):
        return self._array.parameters

    def parameter(self, key):
        return self._array.parameter(key)

    def purelist_parameter(self, key):
        return self._array.purelist_parameter(key)

    @property
    def purelist_isregular(self):
        return self._array.purelist_isregular

    @property
    def purelist_depth(self):
        return 0

    @property
    def minmax_depth(self):
        mindepth, maxdepth = self._array.minmax_depth
        return mindepth - 1, maxdepth - 1

    @property
    def branch_depth(self):
        branch, depth = self._array.branch_depth
        return branch, depth - 1

    def __getitem__(self, where):
        if ak._util.isint(where):
            raise IndexError("scalar Record cannot be sliced by an integer")

        elif isinstance(where, slice):
            raise IndexError("scalar Record cannot be sliced by a range slice (`:`)")

        elif ak._util.isstr(where):
            return self._getitem_field(where)

        elif where is np.newaxis:
            raise IndexError("scalar Record cannot be sliced by np.newaxis (`None`)")

        elif where is Ellipsis:
            raise IndexError("scalar Record cannot be sliced by an ellipsis (`...`)")

        elif isinstance(where, tuple) and len(where) == 0:
            return self

        elif isinstance(where, tuple) and len(where) == 1:
            return self.__getitem__(where[0])

        elif isinstance(where, tuple) and ak._util.isstr(where[0]):
            return self._getitem_field(where[0]).__getitem__(where[1:])

        elif isinstance(where, ak.highlevel.Array):
            raise IndexError("scalar Record cannot be sliced by an array")

        elif isinstance(where, ak.layout.Content):
            raise IndexError("scalar Record cannot be sliced by an array")

        elif isinstance(where, Content):
            raise IndexError("scalar Record cannot be sliced by an array")

        elif isinstance(where, Iterable) and all(ak._util.isstr(x) for x in where):
            return self._getitem_fields(where)

        elif isinstance(where, Iterable):
            raise IndexError("scalar Record cannot be sliced by an array")

        else:
            raise TypeError(
                "only field name (str) or names (non-tuple iterable of str) "
                "are valid indices for slicing a scalar record, not\n\n    "
                + repr(where)
            )

    def _getitem_field(self, where):
        return self._array._getitem_field(where)._getitem_at(self._at)

    def _getitem_fields(self, where):
        return self._array._getitem_fields(where)._getitem_at(self._at)

    def packed(self):
        if self._array.length == 1:
            return Record(self._array.packed(), self._at)
        else:
            return Record(self._array[self._at : self._at + 1].packed(), 0)

    def to_list(self, behavior=None):
        cls = ak._v2._util.recordclass(self._array, behavior)
        if cls is not ak._v2.highlevel.Record:
            return cls(self)

        return self._array[self._at : self._at + 1].to_list(behavior)[0]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import urllib.request
import urllib.error

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


@pytest.mark.skipif(not ak._util.win, reason="need Windows to test this case")
def test():
    url = "https://raw.githubusercontent.com/Chicago/osd-bike-routes/5f556dc/data/Bikeroutes.geojson"
    try:
        bikeroutes_json = urllib.request.urlopen(url).read()
    except urllib.error.URLError:
        pytest.skip(msg="couldn't download sample dataset")

    # This shouldn't fail (see #1084)
    ak.from_json(bikeroutes_json)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    builder = ak.layout.LayoutBuilder64(
        """
    {
        "form_key": "root",
        "class": "RecordArray",
        "contents": {
            "u": {
                "form_key": "u",
                "class": "ListOffsetArray64",
                "offsets": "i64",
                "content": {
                    "form_key": "u-content",
                    "class": "RecordArray",
                    "contents": {
                        "i": {
                            "class": "NumpyArray",
                            "primitive": "int64",
                            "form_key": "i"
                        },
                        "j": {
                            "class": "ListOffsetArray64",
                            "offsets": "i64",
                            "content": "int64",
                            "form_key": "j"
                        }
                    }
                }
            },
            "v": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "v"
            },
            "w": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "w"
            },
            "x": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "x"
            }
        }
    }
    """
    )

    builder.begin_list()  # u

    builder.int64(1)  # i
    builder.begin_list()  # j
    builder.int64(9)
    builder.int64(8)
    builder.int64(7)
    builder.end_list()  # j

    builder.end_list()  # u

    builder.int64(2)  # v
    builder.int64(3)  # w
    builder.int64(4)  # x

    builder.begin_list()  # u

    builder.int64(1)  # i
    builder.begin_list()  # j
    builder.int64(9)
    builder.int64(8)
    builder.int64(7)
    builder.end_list()  # j

    builder.end_list()  # u

    builder.int64(2)  # v
    builder.int64(3)  # w
    builder.int64(4)  # x

    assert ak.to_list(builder.snapshot()) == [
        {"u": [{"i": 1, "j": [9, 8, 7]}], "v": 2, "w": 3, "x": 4},
        {"u": [{"i": 1, "j": [9, 8, 7]}], "v": 2, "w": 3, "x": 4},
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


# @ak._v2._connect.numpy.implements("nan_to_num")
def nan_to_num(
    array, copy=True, nan=0.0, posinf=None, neginf=None, highlevel=True, behavior=None
):
    raise NotImplementedError


#     """
#     Args:
#         array: Array whose `NaN` values should be converted to a number.
#         copy (bool): Ignored (Awkward Arrays are immutable).
#         nan (int, float, broadcastable array): Value to be used to fill `NaN` values.
#         posinf (None, int, float, broadcastable array): Value to be used to fill positive infinity
#             values. If None, positive infinities are replaced with a very large number.
#         neginf (None, int, float, broadcastable array): Value to be used to fill negative infinity
#             values. If None, negative infinities are replaced with a very small number.
#         highlevel (bool): If True, return an #ak.Array; otherwise, return
#             a low-level #ak.layout.Content subclass.
#         behavior (None or dict): Custom #ak.behavior for the output array, if
#             high-level.

#     Implements [np.nan_to_num](https://numpy.org/doc/stable/reference/generated/numpy.nan_to_num.html)
#     for Awkward Arrays.
#     """
#     behavior = ak._util.behaviorof(array, behavior=behavior)

#     broadcasting_ids = {}
#     broadcasting = []

#     layout = ak.operations.convert.to_layout(array)
#     broadcasting.append(layout)

#     nan_layout = ak.operations.convert.to_layout(nan, allow_other=True)
#     if isinstance(nan_layout, ak.layout.Content):
#         broadcasting_ids[id(nan)] = len(broadcasting)
#         broadcasting.append(nan_layout)

#     posinf_layout = ak.operations.convert.to_layout(posinf, allow_other=True)
#     if isinstance(posinf_layout, ak.layout.Content):
#         broadcasting_ids[id(posinf)] = len(broadcasting)
#         broadcasting.append(posinf_layout)

#     neginf_layout = ak.operations.convert.to_layout(neginf, allow_other=True)
#     if isinstance(neginf_layout, ak.layout.Content):
#         broadcasting_ids[id(neginf)] = len(broadcasting)
#         broadcasting.append(neginf_layout)

#     nplike = ak.nplike.of(layout)

#     if len(broadcasting) == 1:

#         def getfunction(layout):
#             if isinstance(layout, ak.layout.NumpyArray):
#                 return lambda: ak.layout.NumpyArray(
#                     nplike.nan_to_num(
#                         nplike.asarray(layout),
#                         nan=nan,
#                         posinf=posinf,
#                         neginf=neginf,
#                     )
#                 )
#             else:
#                 return None

#         out = ak._util.recursively_apply(
#             layout, getfunction, pass_depth=False, pass_user=False
#         )

#     else:

#         def getfunction(inputs):
#             if all(isinstance(x, ak.layout.NumpyArray) for x in inputs):
#                 tmp_layout = nplike.asarray(inputs[0])
#                 if id(nan) in broadcasting_ids:
#                     tmp_nan = nplike.asarray(inputs[broadcasting_ids[id(nan)]])
#                 else:
#                     tmp_nan = nan
#                 if id(posinf) in broadcasting_ids:
#                     tmp_posinf = nplike.asarray(inputs[broadcasting_ids[id(posinf)]])
#                 else:
#                     tmp_posinf = posinf
#                 if id(neginf) in broadcasting_ids:
#                     tmp_neginf = nplike.asarray(inputs[broadcasting_ids[id(neginf)]])
#                 else:
#                     tmp_neginf = neginf
#                 return lambda: (ak.layout.NumpyArray(
#                     nplike.nan_to_num(
#                         tmp_layout,
#                         nan=tmp_nan,
#                         posinf=tmp_posinf,
#                         neginf=tmp_neginf,
#                     )
#                 ),)
#             else:
#                 return None

#         out = ak._util.broadcast_and_apply(
#             broadcasting,
#             getfunction,
#             behavior,
#             pass_depth=False,
#         )
#         assert isinstance(out, tuple) and len(out) == 1
#         out = out[0]

#     return ak._util.maybe_wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

to_list = ak._v2.operations.convert.to_list


def test_1d():
    array = ak._v2.operations.convert.from_iter(
        [3.3, 2.2, 5.5, 1.1, 4.4],
        highlevel=False,
    )
    assert to_list(array.argmin(axis=0)) == 3
    assert to_list(array.argmax(axis=0)) == 2
    assert to_list(array.count(axis=0)) == 5
    assert to_list(array.count_nonzero(axis=0)) == 5
    assert to_list(array.sum(axis=0)) == 16.5
    assert to_list(array.prod(axis=0)) == 193.26120000000003
    assert to_list(array.any(axis=0)) is True
    assert to_list(array.all(axis=0)) is True
    assert to_list(array.min(axis=0)) == 1.1
    assert to_list(array.max(axis=0)) == 5.5


def test_2d():
    array = ak._v2.operations.convert.from_iter(
        [
            [3.3, 2.2, 5.5, 1.1, 4.4],
            [4.4, 2.2, 1.1, 3.3, 5.5],
            [2.2, 1.1, 4.4, 3.3, 5.5],
        ],
        highlevel=False,
    )
    assert to_list(array.argmin(axis=0)) == [2, 2, 1, 0, 0]
    assert array.typetracer.argmin(axis=0).form == array.argmin(axis=0).form
    assert to_list(array.argmin(axis=1)) == [3, 2, 1]
    assert array.typetracer.argmin(axis=1).form == array.argmin(axis=1).form


def test_3d():
    array = ak._v2.operations.convert.from_iter(
        [
            [
                [3.3, 2.2, 5.5, 1.1, 4.4],
                [4.4, 2.2, 1.1, 3.3, 5.5],
                [2.2, 1.1, 4.4, 3.3, 5.5],
            ],
            [
                [-3.3, 2.2, -5.5, 1.1, 4.4],
                [4.4, -2.2, 1.1, -3.3, 5.5],
                [2.2, 1.1, 4.4, 3.3, -5.5],
            ],
        ],
        highlevel=False,
    )

    assert to_list(array.argmin(axis=0)) == [
        [1, 0, 1, 0, 0],
        [0, 1, 0, 1, 0],
        [0, 0, 0, 0, 1],
    ]
    assert to_list(array.argmin(axis=1)) == [[2, 2, 1, 0, 0], [0, 1, 0, 1, 2]]
    assert array.typetracer.argmin(axis=1).form == array.argmin(axis=1).form
    assert to_list(array.argmin(axis=2)) == [[3, 2, 1], [2, 3, 4]]
    assert array.typetracer.argmin(axis=2).form == array.argmin(axis=2).form
    assert to_list(array.argmin(axis=-1)) == [[3, 2, 1], [2, 3, 4]]
    assert array.typetracer.argmin(axis=-1).form == array.argmin(axis=-1).form
    assert to_list(array.argmin(axis=-2)) == [[2, 2, 1, 0, 0], [0, 1, 0, 1, 2]]
    assert array.typetracer.argmin(axis=-2).form == array.argmin(axis=-2).form
    assert to_list(array.argmin(axis=-3)) == [
        [1, 0, 1, 0, 0],
        [0, 1, 0, 1, 0],
        [0, 0, 0, 0, 1],
    ]
    assert array.typetracer.argmin(axis=-3).form == array.argmin(axis=-3).form


def test_jagged():
    v2_array = ak._v2.operations.convert.from_iter(
        [[2.2, 1.1, 3.3], [], [4.4, 5.5], [5.5], [-4.4, -5.5, -6.6]], highlevel=False
    )
    assert to_list(v2_array.argmin(axis=1)) == [1, None, 0, 0, 2]
    assert v2_array.typetracer.argmin(axis=1).form == v2_array.argmin(axis=1).form

    index2 = ak._v2.index.Index64(np.array([4, 3, 2, 1, 0], dtype=np.int64))
    v2_array2 = ak._v2.contents.IndexedArray(index2, v2_array)
    assert to_list(v2_array2.argmin(axis=1)) == [2, 0, 0, None, 1]
    assert v2_array2.typetracer.argmin(axis=1).form == v2_array2.argmin(axis=1).form

    index3 = ak._v2.index.Index64(np.array([4, 3, -1, 4, 0], dtype=np.int64))
    v2_array2 = ak._v2.contents.IndexedOptionArray(index3, v2_array)
    assert to_list(v2_array2.argmin(axis=1)) == [2, 0, None, 2, 1]
    assert v2_array2.typetracer.argmin(axis=1).form == v2_array2.argmin(axis=1).form
    assert to_list(v2_array2.argmin(axis=-1)) == [2, 0, None, 2, 1]
    assert v2_array2.typetracer.argmin(axis=-1).form == v2_array2.argmin(axis=-1).form


def test_missing():
    array = ak._v2.operations.convert.from_iter(
        [[[2.2, 1.1, 3.3]], [[]], [None, None, None], [[-4.4, -5.5, -6.6]]],
        highlevel=False,
    )
    assert to_list(array.argmin(axis=2)) == [[1], [None], [None, None, None], [2]]
    assert array.typetracer.argmin(axis=2).form == array.argmin(axis=2).form



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


# @ak._v2._connect.numpy.implements("copy")
def copy(array):
    raise NotImplementedError


#     """
#     Returns a deep copy of the array (no memory shared with original).

#     This is identical to `np.copy` and `copy.deepcopy`.

#     It's only useful to explicitly copy an array if you're going to change it
#     in-place. This doesn't come up often because Awkward Arrays are immutable.
#     That is to say, the Awkward Array library doesn't have any operations that
#     change an array in-place, but the data in the array might be owned by another
#     library that can change it in-place.

#     For example, if the array comes from NumPy:

#         >>> underlying_array = np.array([1.1, 2.2, 3.3, 4.4, 5.5])
#         >>> wrapper = ak.Array(underlying_array)
#         >>> duplicate = ak.copy(wrapper)
#         >>> underlying_array[2] = 123
#         >>> underlying_array
#         array([  1.1,   2.2, 123. ,   4.4,   5.5])
#         >>> wrapper
#         <Array [1.1, 2.2, 123, 4.4, 5.5] type='5 * float64'>
#         >>> duplicate
#         <Array [1.1, 2.2, 3.3, 4.4, 5.5] type='5 * float64'>

#     There is an exception to this rule: you can add fields to records in an
#     #ak.Array in-place. However, this changes the #ak.Array wrapper without
#     affecting the underlying layout data (it *replaces* its layout), so a
#     shallow copy will do:

#         >>> original = ak.Array([{"x": 1}, {"x": 2}, {"x": 3}])
#         >>> shallow_copy = copy.copy(original)
#         >>> shallow_copy["y"] = original.x**2
#         >>> shallow_copy
#         <Array [{x: 1, y: 1}, ... y: 4}, {x: 3, y: 9}] type='3 * {"x": int64, "y": int64}'>
#         >>> original
#         <Array [{x: 1}, {x: 2}, {x: 3}] type='3 * {"x": int64}'>

#     This is key to Awkward Array's efficiency (memory and speed): operations that
#     only change part of a structure re-use pieces from the original ("structural
#     sharing"). Changing data in-place would result in many surprising long-distance
#     changes, so we don't support it. However, an #ak.Array's data might come from
#     a mutable third-party library, so this function allows you to make a true copy.
#     """
#     layout = ak._v2.operations.convert.to_layout(
#         array,
#         allow_record=True,
#         allow_other=False,
#     )
#     return ak._v2._util.wrap(layout.deep_copy(), ak._v2._util.behaviorof(array))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def to_jax(array):
    raise NotImplementedError


#     """
#     Converts `array` (many types supported) into a JAX array, if possible.

#     If the data are numerical and regular (nested lists have equal lengths
#     in each dimension, as described by the #type), they can be losslessly
#     converted to a CuPy array and this function returns without an error.

#     Otherwise, the function raises an error.

#     If `array` is a scalar, it is converted into a JAX scalar.

#     See also #ak.to_cupy, #ak.from_jax and #ak.to_numpy.
#     """
#     try:
#         import jax
#     except ImportError:
#         raise ImportError(
#             """to use {0}, you must install jax:

#                 pip install jax jaxlib
#             """
#         )

#     if isinstance(array, (bool, numbers.Number)):
#         return jax.numpy.array([array])[0]

#     elif isinstance(array, jax.numpy.ndarray):
#         return array

#     elif isinstance(array, np.ndarray):
#         return jax.numpy.asarray(array)

#     elif isinstance(array, ak._v2.highlevel.Array):
#         return to_jax(array.layout)

#     elif isinstance(array, ak._v2.highlevel.Record):
#         raise ValueError(
#             "JAX does not support record structures"
#
#         )

#     elif isinstance(array, ak._v2.highlevel.ArrayBuilder):
#         return to_jax(array.snapshot().layout)

#     elif isinstance(array, ak.layout.ArrayBuilder):
#         return to_jax(array.snapshot())

#     elif (
#         ak._v2.operations.describe.parameters(array).get("__array__") == "bytestring"
#         or ak._v2.operations.describe.parameters(array).get("__array__") == "string"
#     ):
#         raise ValueError(
#             "JAX does not support arrays of strings"
#
#         )

#     elif isinstance(array, ak.partition.PartitionedArray):   # NO PARTITIONED ARRAY
#         return jax.numpy.concatenate([to_jax(x) for x in array.partitions])

#     elif isinstance(array, ak._v2._util.virtualtypes):
#         return to_jax(array.array)

#     elif isinstance(array, ak._v2._util.unknowntypes):
#         return jax.numpy.array([])

#     elif isinstance(array, ak._v2._util.indexedtypes):
#         return to_jax(array.project())

#     elif isinstance(array, ak._v2._util.uniontypes):
#         array = array.simplify()
#         if isinstance(array, ak._v2._util.uniontypes):
#             raise ValueError(
#                 "cannot convert {0} into jax.numpy.array".format(array)
#
#             )
#         return to_jax(array)

#     elif isinstance(array, ak._v2.contents.UnmaskedArray):
#         return to_jax(array.content)

#     elif isinstance(array, ak._v2._util.optiontypes):
#         content = to_jax(array.project())

#         shape = list(content.shape)
#         shape[0] = len(array)
#         mask0 = jax.numpy.asarray(array.bytemask()).view(np.bool_)
#         if mask0.any():
#             raise ValueError(
#                 "JAX does not support masked arrays"
#
#             )
#         else:
#             return content

#     elif isinstance(array, ak._v2.contents.RegularArray):
#         out = to_jax(array.content)
#         head, tail = out.shape[0], out.shape[1:]
#         shape = (head // array.size, array.size) + tail
#         return out[: shape[0] * array.size].reshape(shape)

#     elif isinstance(array, ak._v2._util.listtypes):
#         return to_jax(array.toRegularArray())

#     elif isinstance(array, ak._v2._util.recordtypes):
#         raise ValueError(
#             "JAX does not support record structures"
#
#         )

#     elif isinstance(array, ak._v2.contents.NumpyArray):
#         return array.to_jax()

#     elif isinstance(array, ak._v2.contents.Content):
#         raise AssertionError(
#             "unrecognized Content type: {0}".format(type(array))
#
#         )

#     elif isinstance(array, Iterable):
#         return jax.numpy.asarray(array)

#     else:
#         raise ValueError(
#             "cannot convert {0} into jax.numpy.array".format(array)
#
#         )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    a = ak.layout.RegularArray(ak.layout.EmptyArray(), size=0, zeros_length=3)
    assert ak.to_numpy(a).shape == (3, 0)
    assert ak.to_numpy(ak.Array([[], [], []])).shape == (3, 0)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

pytest.importorskip("pyarrow")


def test():
    a = ak._v2.highlevel.Array(
        ak._v2.contents.RecordArray(
            [
                ak._v2.contents.NumpyArray(np.array([1.1, 2.2, 3.3])),
                ak._v2.contents.NumpyArray(np.array([1, 2, 3])),
            ],
            ["x", "y"],
            parameters={"__record__": "Hello"},
        )
    )
    assert ak._v2.operations.convert.from_arrow(
        ak._v2.operations.convert.to_arrow_table(a)
    ).type.content.parameters == {"__record__": "Hello"}



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


# @ak._v2._connect.numpy.implements("sort")
def sort(array, axis=-1, ascending=True, stable=True, highlevel=True, behavior=None):

    """
        Args:
            array: Data to sort, possibly within nested lists.
            axis (int): The dimension at which this operation is applied. The
                outermost dimension is `0`, followed by `1`, etc., and negative
                values count backward from the innermost: `-1` is the innermost
                dimension, `-2` is the next level up, etc.
            ascending (bool): If True, the first value in each sorted group
                will be smallest, the last value largest; if False, the order
                is from largest to smallest.
            stable (bool): If True, use a stable sorting algorithm (introsort:
                a hybrid of quicksort, heapsort, and insertion sort); if False,
                use a sorting algorithm that is not guaranteed to be stable
                (heapsort).
            highlevel (bool): If True, return an #ak.Array; otherwise, return
                a low-level #ak.layout.Content subclass.
            behavior (None or dict): Custom #ak.behavior for the output array, if
                high-level.

    #     For example,

    #         >>> ak.sort(ak.Array([[7, 5, 7], [], [2], [8, 2]]))
    #         <Array [[5, 7, 7], [], [2], [2, 8]] type='4 * var * int64'>
    #"""
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )
    out = layout.sort(axis, ascending, stable)
    return ak._v2._util.wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

pytest.importorskip("pyarrow.parquet")


def test_parquet():
    empty = ak.from_parquet("tests/samples/zero-record-batches.parquet")
    assert isinstance(empty, ak.Array)
    assert len(empty) == 0
    assert str(empty.type) == "0 * {}"


def test_concatenate():
    one = ak.Array(ak.layout.RecordArray([], [], length=0))
    two = ak.Array(ak.layout.RecordArray([], [], length=0))
    assert len(ak.concatenate([one, two])) == 0
    assert len(ak.concatenate([one])) == 0

    one = ak.Array(ak.layout.RecordArray([], [], length=3))
    two = ak.Array(ak.layout.RecordArray([], [], length=5))
    assert len(ak.concatenate([one, two])) == 8
    assert len(ak.concatenate([one])) == 3



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

to_list = ak._v2.operations.convert.to_list

primes = [x for x in range(2, 1000) if all(x % n != 0 for n in range(2, x))]


def test():
    data = ak._v2.highlevel.Array([[1, 3, 5, 4, 2], [], [2, 3, 1], [5]]).layout
    assert to_list(data.min(axis=1, initial=4)) == [1, None, 1, 4]
    assert (
        data.typetracer.min(axis=1, initial=4).form == data.min(axis=1, initial=4).form
    )
    assert to_list(data.min(axis=1)) == [1, None, 1, 5]
    assert data.typetracer.min(axis=1).form == data.min(axis=1).form
    assert to_list(data.max(axis=1, initial=4)) == [5, None, 4, 5]
    assert (
        data.typetracer.max(axis=1, initial=4).form == data.max(axis=1, initial=4).form
    )
    assert to_list(data.max(axis=1)) == [5, None, 3, 5]
    assert data.typetracer.max(axis=1).form == data.max(axis=1).form

    data = ak._v2.highlevel.Array(
        [[1.1, 3.3, 5.5, 4.4, 2.2], [], [2.2, 3.3, 1.1], [5.5]]
    ).layout
    assert to_list(data.min(axis=1, initial=4)) == [1.1, None, 1.1, 4]
    assert (
        data.typetracer.min(axis=1, initial=4).form == data.min(axis=1, initial=4).form
    )
    assert to_list(data.min(axis=1)) == [1.1, None, 1.1, 5.5]
    assert data.typetracer.min(axis=1).form == data.min(axis=1).form
    assert to_list(data.max(axis=1, initial=4)) == [5.5, None, 4, 5.5]
    assert (
        data.typetracer.max(axis=1, initial=4).form == data.max(axis=1, initial=4).form
    )
    assert to_list(data.max(axis=1)) == [5.5, None, 3.3, 5.5]
    assert data.typetracer.max(axis=1).form == data.max(axis=1).form

    data = ak._v2.contents.NumpyArray(np.array(primes[2 : 2 * 3 * 5], dtype=np.int32))
    assert to_list(data.min(axis=-1, initial=4)) == 4
    assert to_list(data.min(axis=-1)) == 5
    assert to_list(data.max(axis=-1, initial=114)) == 114
    assert to_list(data.max(axis=-1)) == 113

    data = ak._v2.contents.NumpyArray(np.array(primes[2 : 2 * 3 * 5], dtype=np.uint64))
    assert to_list(data.min(axis=-1, initial=4)) == 4
    assert to_list(data.min(axis=-1)) == 5
    assert to_list(data.max(axis=-1, initial=114)) == 114
    assert to_list(data.max(axis=-1)) == 113

    data = ak._v2.contents.NumpyArray(np.array(primes[2 : 2 * 3 * 5], dtype=np.uint32))
    assert to_list(data.min(axis=-1, initial=4)) == 4
    assert to_list(data.min(axis=-1)) == 5
    assert to_list(data.max(axis=-1, initial=114)) == 114
    assert to_list(data.max(axis=-1)) == 113


def test_date_time():
    numpy_array = np.array(
        ["2020-07-27T10:41:11", "2019-01-01", "2020-01-01"], "datetime64[s]"
    )

    array = ak._v2.highlevel.Array(numpy_array)
    assert str(array.type) == "3 * datetime64[s]"
    assert array.tolist() == [
        np.datetime64("2020-07-27T10:41:11"),
        np.datetime64("2019-01-01T00:00:00"),
        np.datetime64("2020-01-01T00:00:00"),
    ]

    array = array.layout
    assert to_list(array) == [
        np.datetime64("2020-07-27T10:41:11"),
        np.datetime64("2019-01-01T00:00:00"),
        np.datetime64("2020-01-01T00:00:00"),
    ]

    assert array.max() == numpy_array[0]
    assert array.min() == numpy_array[1]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    a = ak.Array([{"this": 100}])
    b = ak.Array([{"this": 90, "that": 100}])
    c = ak.concatenate((a, b))

    with pytest.raises(ValueError):
        ak.unzip(c)

    a = ak.Array([{"this": 100}])
    b = ak.Array([{"this": 90}])
    c = ak.concatenate((a, b))

    (tmp,) = ak.unzip(c)

    assert tmp.tolist() == [100, 90]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import copy

import awkward as ak
from awkward._v2.index import Index
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.contents.listoffsetarray import ListOffsetArray
from awkward._v2.forms.listform import ListForm
from awkward._v2.forms.form import _parameters_equal

np = ak.nplike.NumpyMetadata.instance()


class ListArray(Content):
    is_ListType = True

    def __init__(
        self, starts, stops, content, identifier=None, parameters=None, nplike=None
    ):
        if not isinstance(starts, Index) and starts.dtype in (
            np.dtype(np.int32),
            np.dtype(np.uint32),
            np.dtype(np.int64),
        ):
            raise TypeError(
                "{} 'starts' must be an Index with dtype in (int32, uint32, int64), "
                "not {}".format(type(self).__name__, repr(starts))
            )
        if not (isinstance(stops, Index) and starts.dtype == stops.dtype):
            raise TypeError(
                "{} 'stops' must be an Index with the same dtype as 'starts' ({}), "
                "not {}".format(type(self).__name__, repr(starts.dtype), repr(stops))
            )
        if not isinstance(content, Content):
            raise TypeError(
                "{} 'content' must be a Content subtype, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if (
            starts.nplike.known_shape
            and stops.nplike.known_shape
            and starts.length > stops.length
        ):
            raise ValueError(
                "{} len(starts) ({}) must be <= len(stops) ({})".format(
                    type(self).__name__, starts.length, stops.length
                )
            )
        if nplike is None:
            nplike = content.nplike
        if nplike is None:
            nplike = starts.nplike

        self._starts = starts
        self._stops = stops
        self._content = content
        self._init(identifier, parameters, nplike)

    @property
    def starts(self):
        return self._starts

    @property
    def stops(self):
        return self._stops

    @property
    def content(self):
        return self._content

    Form = ListForm

    def _form_with_key(self, getkey):
        form_key = getkey(self)
        return self.Form(
            self._starts.form,
            self._stops.form,
            self._content._form_with_key(getkey),
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=form_key,
        )

    def _to_buffers(self, form, getkey, container, nplike):
        assert isinstance(form, self.Form)
        key1 = getkey(self, form, "starts")
        key2 = getkey(self, form, "stops")
        container[key1] = ak._v2._util.little_endian(self._starts.raw(nplike))
        container[key2] = ak._v2._util.little_endian(self._stops.raw(nplike))
        self._content._to_buffers(form.content, getkey, container, nplike)

    @property
    def typetracer(self):
        tt = ak._v2._typetracer.TypeTracer.instance()
        return ListArray(
            ak._v2.index.Index(self._starts.raw(tt)),
            ak._v2.index.Index(self._stops.raw(tt)),
            self._content.typetracer,
            self._typetracer_identifier(),
            self._parameters,
            tt,
        )

    @property
    def length(self):
        return self._starts.length

    def _forget_length(self):
        return ListArray(
            self._starts.forget_length(),
            self._stops,
            self._content,
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<ListArray len="]
        out.append(repr(str(self.length)))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._starts._repr(indent + "    ", "<starts>", "</starts>\n"))
        out.append(self._stops._repr(indent + "    ", "<stops>", "</stops>\n"))
        out.append(self._content._repr(indent + "    ", "<content>", "</content>\n"))
        out.append(indent + "</ListArray>")
        out.append(post)
        return "".join(out)

    def merge_parameters(self, parameters):
        return ListArray(
            self._starts,
            self._stops,
            self._content,
            self._identifier,
            ak._v2._util.merge_parameters(self._parameters, parameters),
            self._nplike,
        )

    def toListOffsetArray64(self, start_at_zero=False):
        starts = self._starts.data
        stops = self._stops.data
        if self._nplike.array_equal(starts[1:], stops[:-1]):
            offsets = self._nplike.empty(starts.shape[0] + 1, dtype=starts.dtype)
            if offsets.shape[0] == 1:
                offsets[0] = 0
            else:
                offsets[:-1] = starts
                offsets[-1] = stops[-1]
            return ListOffsetArray(
                ak._v2.index.Index(offsets),
                self._content,
                self._identifier,
                self._parameters,
                self._nplike,
            ).toListOffsetArray64(start_at_zero=start_at_zero)

        else:
            offsets = self._compact_offsets64(start_at_zero)
            return self._broadcast_tooffsets64(offsets)

    def toRegularArray(self):
        offsets = self._compact_offsets64(True)
        return self._broadcast_tooffsets64(offsets).toRegularArray()

    def _getitem_nothing(self):
        return self._content._getitem_range(slice(0, 0))

    def _getitem_at(self, where):
        if not self._nplike.known_data:
            return self._content._getitem_range(slice(0, 0))

        if where < 0:
            where += self.length
        if not (0 <= where < self.length) and self._nplike.known_shape:
            raise NestedIndexError(self, where)
        start, stop = self._starts[where], self._stops[where]
        return self._content._getitem_range(slice(start, stop))

    def _getitem_range(self, where):
        if not self._nplike.known_shape:
            return self

        start, stop, step = where.indices(self.length)
        assert step == 1
        return ListArray(
            self._starts[start:stop],
            self._stops[start:stop],
            self._content,
            self._range_identifier(start, stop),
            self._parameters,
            self._nplike,
        )

    def _getitem_field(self, where, only_fields=()):
        return ListArray(
            self._starts,
            self._stops,
            self._content._getitem_field(where, only_fields),
            self._field_identifier(where),
            None,
            self._nplike,
        )

    def _getitem_fields(self, where, only_fields=()):
        return ListArray(
            self._starts,
            self._stops,
            self._content._getitem_fields(where, only_fields),
            self._fields_identifier(where),
            None,
            self._nplike,
        )

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)

        try:
            nextstarts = self._starts[carry.data]
            nextstops = self._stops[: self._starts.length][carry.data]
        except IndexError as err:
            if issubclass(exception, NestedIndexError):
                raise exception(self, carry.data, str(err))
            else:
                raise exception(str(err))

        return ListArray(
            nextstarts,
            nextstops,
            self._content,
            self._carry_identifier(carry, exception),
            self._parameters,
            self._nplike,
        )

    def _compact_offsets64(self, start_at_zero):
        starts_len = self._starts.length
        out = ak._v2.index.Index64.empty(starts_len + 1, self._nplike)
        assert (
            out.nplike is self._nplike
            and self._starts.nplike is self._nplike
            and self._stops.nplike is self._nplike
        )
        self._handle_error(
            self._nplike[
                "awkward_ListArray_compact_offsets",
                out.dtype.type,
                self._starts.dtype.type,
                self._stops.dtype.type,
            ](
                out.data,
                self._starts.data,
                self._stops.data,
                starts_len,
            )
        )
        return out

    def _broadcast_tooffsets64(self, offsets):
        return ListOffsetArray._broadcast_tooffsets64(self, offsets)

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        slicestarts = slicestarts._to_nplike(self.nplike)
        slicestops = slicestops._to_nplike(self.nplike)
        if self._nplike.known_shape and slicestarts.length != self.length:
            raise NestedIndexError(
                self,
                ak._v2.contents.ListArray(
                    slicestarts, slicestops, slicecontent, None, None, self._nplike
                ),
                "cannot fit jagged slice with length {} into {} of size {}".format(
                    slicestarts.length, type(self).__name__, self.length
                ),
            )

        if isinstance(slicecontent, ak._v2.contents.listoffsetarray.ListOffsetArray):
            outoffsets = ak._v2.index.Index64.empty(
                slicestarts.length + 1, self._nplike
            )
            assert (
                outoffsets.nplike is self._nplike
                and slicestarts.nplike is self._nplike
                and slicestops.nplike is self._nplike
                and self._starts.nplike is self._nplike
                and self._stops.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_ListArray_getitem_jagged_descend",
                    outoffsets.dtype.type,
                    slicestarts.dtype.type,
                    slicestops.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    outoffsets.data,
                    slicestarts.data,
                    slicestops.data,
                    slicestarts.length,
                    self._starts.data,
                    self._stops.data,
                )
            )

            asListOffsetArray64 = self.toListOffsetArray64(True)
            next_content = asListOffsetArray64._content

            sliceoffsets = ak._v2.index.Index64(slicecontent._offsets)

            outcontent = next_content._getitem_next_jagged(
                sliceoffsets[:-1], sliceoffsets[1:], slicecontent._content, tail
            )

            return ak._v2.contents.listoffsetarray.ListOffsetArray(
                outoffsets, outcontent, None, self._parameters, self._nplike
            )

        elif isinstance(slicecontent, ak._v2.contents.numpyarray.NumpyArray):
            carrylen = ak._v2.index.Index64.empty(1, self._nplike)
            assert (
                carrylen.nplike is self._nplike
                and slicestarts.nplike is self._nplike
                and slicestops.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_ListArray_getitem_jagged_carrylen",
                    carrylen.dtype.type,
                    slicestarts.dtype.type,
                    slicestops.dtype.type,
                ](
                    carrylen.data,
                    slicestarts.data,
                    slicestops.data,
                    slicestarts.length,
                )
            )
            sliceindex = ak._v2.index.Index64(slicecontent._data)
            outoffsets = ak._v2.index.Index64.empty(
                slicestarts.length + 1, self._nplike
            )
            nextcarry = ak._v2.index.Index64.empty(carrylen[0], self._nplike)

            assert (
                outoffsets.nplike is self._nplike
                and nextcarry.nplike is self._nplike
                and slicestarts.nplike is self._nplike
                and slicestops.nplike is self._nplike
                and sliceindex.nplike is self._nplike
                and self._starts.nplike is self._nplike
                and self._stops.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_ListArray_getitem_jagged_apply",
                    outoffsets.dtype.type,
                    nextcarry.dtype.type,
                    slicestarts.dtype.type,
                    slicestops.dtype.type,
                    sliceindex.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    outoffsets.data,
                    nextcarry.data,
                    slicestarts.data,
                    slicestops.data,
                    slicestarts.length,
                    sliceindex.data,
                    sliceindex.length,
                    self._starts.data,
                    self._stops.data,
                    self._content.length,
                )
            )

            nextcontent = self._content._carry(nextcarry, True, NestedIndexError)
            nexthead, nexttail = ak._v2._slicing.headtail(tail)
            outcontent = nextcontent._getitem_next(nexthead, nexttail, None)

            return ak._v2.contents.listoffsetarray.ListOffsetArray(
                outoffsets, outcontent, None, None, self._nplike
            )

        elif isinstance(
            slicecontent, ak._v2.contents.indexedoptionarray.IndexedOptionArray
        ):
            if self._nplike.known_shape and self._starts.length < slicestarts.length:
                raise NestedIndexError(
                    self,
                    ak._v2.contents.ListArray(
                        slicestarts, slicestops, slicecontent, None, None, self._nplike
                    ),
                    "jagged slice length differs from array length",
                )

            missing = ak._v2.index.Index64(slicecontent._index)
            numvalid = ak._v2.index.Index64.empty(1, self._nplike)
            assert (
                numvalid.nplike is self._nplike
                and slicestarts.nplike is self._nplike
                and slicestops.nplike is self._nplike
                and missing.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_ListArray_getitem_jagged_numvalid",
                    numvalid.dtype.type,
                    slicestarts.dtype.type,
                    slicestops.dtype.type,
                    missing.dtype.type,
                ](
                    numvalid.data,
                    slicestarts.data,
                    slicestops.data,
                    slicestarts.length,
                    missing.data,
                    missing.length,
                )
            )

            nextcarry = ak._v2.index.Index64.empty(numvalid[0], self._nplike)

            smalloffsets = ak._v2.index.Index64.empty(
                slicestarts.length + 1, self._nplike
            )
            largeoffsets = ak._v2.index.Index64.empty(
                slicestarts.length + 1, self._nplike
            )

            assert (
                nextcarry.nplike is self._nplike
                and smalloffsets.nplike is self._nplike
                and largeoffsets.nplike is self._nplike
                and slicestarts.nplike is self._nplike
                and slicestops.nplike is self._nplike
                and missing.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_ListArray_getitem_jagged_shrink",
                    nextcarry.dtype.type,
                    smalloffsets.dtype.type,
                    largeoffsets.dtype.type,
                    slicestarts.dtype.type,
                    slicestops.dtype.type,
                    missing.dtype.type,
                ](
                    nextcarry.data,
                    smalloffsets.data,
                    largeoffsets.data,
                    slicestarts.data,
                    slicestops.data,
                    slicestarts.length,
                    missing.data,
                )
            )

            if isinstance(
                slicecontent._content,
                ak._v2.contents.listoffsetarray.ListOffsetArray,
            ):
                nextcontent = self._content._carry(nextcarry, True, NestedIndexError)
                next = ak._v2.contents.listoffsetarray.ListOffsetArray(
                    smalloffsets, nextcontent, None, self._parameters, self._nplike
                )
                out = next._getitem_next_jagged(
                    smalloffsets[:-1], smalloffsets[1:], slicecontent._content, tail
                )

            else:
                out = self._getitem_next_jagged(
                    smalloffsets[:-1], smalloffsets[1:], slicecontent._content, tail
                )

            if isinstance(out, ak._v2.contents.listoffsetarray.ListOffsetArray):
                content = out._content
                if largeoffsets.nplike.known_data:
                    missing_trim = missing[0 : largeoffsets[-1]]
                else:
                    missing_trim = missing
                indexedoptionarray = (
                    ak._v2.contents.indexedoptionarray.IndexedOptionArray(
                        missing_trim, content, None, self._parameters, self._nplike
                    )
                )
                if isinstance(self._nplike, ak._v2._typetracer.TypeTracer):
                    indexedoptionarray = indexedoptionarray.typetracer
                return ak._v2.contents.listoffsetarray.ListOffsetArray(
                    largeoffsets,
                    indexedoptionarray.simplify_optiontype(),
                    None,
                    self._parameters,
                    self._nplike,
                )
            else:
                raise AssertionError(
                    "expected ListOffsetArray from ListArray._getitem_next_jagged, got {}".format(
                        type(out).__name__
                    )
                )

        elif isinstance(slicecontent, ak._v2.contents.emptyarray.EmptyArray):
            return self

        else:
            raise AssertionError(
                "expected Index/IndexedOptionArray/ListOffsetArray in ListArray._getitem_next_jagged, got {}".format(
                    type(slicecontent).__name__
                )
            )

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif isinstance(head, int):
            assert advanced is None
            nexthead, nexttail = ak._v2._slicing.headtail(tail)
            lenstarts = self._starts.length
            nextcarry = ak._v2.index.Index64.empty(lenstarts, self._nplike)
            assert (
                nextcarry.nplike is self._nplike
                and self._starts.nplike is self._nplike
                and self._stops.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_ListArray_getitem_next_at",
                    nextcarry.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    nextcarry.data,
                    self._starts.data,
                    self._stops.data,
                    lenstarts,
                    head,
                )
            )
            nextcontent = self._content._carry(nextcarry, True, NestedIndexError)
            return nextcontent._getitem_next(nexthead, nexttail, advanced)

        elif isinstance(head, slice):
            lenstarts = self._starts.length

            nexthead, nexttail = ak._v2._slicing.headtail(tail)

            start, stop, step = head.start, head.stop, head.step

            step = 1 if step is None else step
            start = ak._util.kSliceNone if start is None else start
            stop = ak._util.kSliceNone if stop is None else stop

            if self._nplike.known_shape:
                carrylength = ak._v2.index.Index64.empty(1, self._nplike)
                assert (
                    carrylength.nplike is self._nplike
                    and self._starts.nplike is self._nplike
                    and self._stops.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_ListArray_getitem_next_range_carrylength",
                        carrylength.dtype.type,
                        self._starts.dtype.type,
                        self._stops.dtype.type,
                    ](
                        carrylength.data,
                        self._starts.data,
                        self._stops.data,
                        lenstarts,
                        start,
                        stop,
                        step,
                    )
                )
                nextcarry = ak._v2.index.Index64.empty(carrylength[0], self._nplike)
            else:
                nextcarry = ak._v2.index.Index64.empty(
                    ak._v2._typetracer.UnknownLength, self._nplike
                )

            if self._starts.dtype == "int64":
                nextoffsets = ak._v2.index.Index64.empty(lenstarts + 1, self._nplike)
            elif self._starts.dtype == "int32":
                nextoffsets = ak._v2.index.Index32.empty(lenstarts + 1, self._nplike)
            elif self._starts.dtype == "uint32":
                nextoffsets = ak._v2.index.IndexU32.empty(lenstarts + 1, self._nplike)

            assert (
                nextoffsets.nplike is self._nplike
                and nextcarry.nplike is self._nplike
                and self._starts.nplike is self._nplike
                and self._stops.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_ListArray_getitem_next_range",
                    nextoffsets.dtype.type,
                    nextcarry.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    nextoffsets.data,
                    nextcarry.data,
                    self._starts.data,
                    self._stops.data,
                    lenstarts,
                    start,
                    stop,
                    step,
                )
            )

            nextcontent = self._content._carry(nextcarry, True, NestedIndexError)

            if advanced is None or advanced.length == 0:
                return ak._v2.contents.listoffsetarray.ListOffsetArray(
                    nextoffsets,
                    nextcontent._getitem_next(nexthead, nexttail, advanced),
                    self._identifier,
                    self._parameters,
                    self._nplike,
                )
            else:
                if self._nplike.known_shape:
                    total = ak._v2.index.Index64.empty(1, self._nplike)
                    assert (
                        total.nplike is self._nplike
                        and nextoffsets.nplike is self._nplike
                    )
                    self._handle_error(
                        self._nplike[
                            "awkward_ListArray_getitem_next_range_counts",
                            total.dtype.type,
                            nextoffsets.dtype.type,
                        ](
                            total.data,
                            nextoffsets.data,
                            lenstarts,
                        )
                    )
                    nextadvanced = ak._v2.index.Index64.empty(total[0], self._nplike)
                else:
                    nextadvanced = ak._v2.index.Index64.empty(
                        ak._v2._typetracer.UnknownLength, self._nplike
                    )
                advanced = advanced._to_nplike(self.nplike)
                assert (
                    nextadvanced.nplike is self._nplike
                    and advanced.nplike is self._nplike
                    and nextoffsets.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_ListArray_getitem_next_range_spreadadvanced",
                        nextadvanced.dtype.type,
                        advanced.dtype.type,
                        nextoffsets.dtype.type,
                    ](
                        nextadvanced.data,
                        advanced.data,
                        nextoffsets.data,
                        lenstarts,
                    )
                )
                return ak._v2.contents.listoffsetarray.ListOffsetArray(
                    nextoffsets,
                    nextcontent._getitem_next(nexthead, nexttail, nextadvanced),
                    self._identifier,
                    self._parameters,
                    self._nplike,
                )

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.index.Index64):
            lenstarts = self._starts.length

            nexthead, nexttail = ak._v2._slicing.headtail(tail)
            flathead = self._nplike.asarray(head.data.reshape(-1))
            regular_flathead = ak._v2.index.Index64(flathead)
            if advanced is None or advanced.length == 0:
                nextcarry = ak._v2.index.Index64.empty(
                    lenstarts * flathead.shape[0], self._nplike
                )
                nextadvanced = ak._v2.index.Index64.empty(
                    lenstarts * flathead.shape[0], self._nplike
                )
                assert (
                    nextcarry.nplike is self._nplike
                    and nextadvanced.nplike is self._nplike
                    and self._starts.nplike is self._nplike
                    and self._stops.nplike is self._nplike
                    and regular_flathead.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_ListArray_getitem_next_array",
                        nextcarry.dtype.type,
                        nextadvanced.dtype.type,
                        self._starts.dtype.type,
                        self._stops.dtype.type,
                        regular_flathead.dtype.type,
                    ](
                        nextcarry.data,
                        nextadvanced.data,
                        self._starts.data,
                        self._stops.data,
                        regular_flathead.data,
                        lenstarts,
                        regular_flathead.length,
                        self._content.length,
                    ),
                    head,
                )
                nextcontent = self._content._carry(nextcarry, True, NestedIndexError)

                out = nextcontent._getitem_next(nexthead, nexttail, nextadvanced)
                if advanced is None:
                    return ak._v2._slicing.getitem_next_array_wrap(
                        out, head.metadata.get("shape", (head.length,))
                    )
                else:
                    return out

            else:
                nextcarry = ak._v2.index.Index64.empty(lenstarts, self._nplike)
                nextadvanced = ak._v2.index.Index64.empty(lenstarts, self._nplike)
                advanced = advanced._to_nplike(self.nplike)
                assert (
                    nextcarry.nplike is self._nplike
                    and nextadvanced.nplike is self._nplike
                    and self._starts.nplike is self._nplike
                    and self._stops.nplike is self._nplike
                    and regular_flathead.nplike is self._nplike
                    and advanced.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_ListArray_getitem_next_array_advanced",
                        nextcarry.dtype.type,
                        nextadvanced.dtype.type,
                        self._starts.dtype.type,
                        self._stops.dtype.type,
                        regular_flathead.dtype.type,
                        advanced.dtype.type,
                    ](
                        nextcarry.data,
                        nextadvanced.data,
                        self._starts.data,
                        self._stops.data,
                        regular_flathead.data,
                        advanced.data,
                        lenstarts,
                        regular_flathead.length,
                        self._content.length,
                    )
                )
                nextcontent = self._content._carry(nextcarry, True, NestedIndexError)

                return nextcontent._getitem_next(nexthead, nexttail, nextadvanced)

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            headlength = head.length
            head = head._to_nplike(self.nplike)
            if advanced is not None:
                raise NestedIndexError(
                    self,
                    head,
                    "cannot mix jagged slice with NumPy-style advanced indexing",
                )
            length = self.length
            singleoffsets = ak._v2.index.Index64(head.offsets.data)
            multistarts = ak._v2.index.Index64.empty(head.length * length, self._nplike)
            multistops = ak._v2.index.Index64.empty(head.length * length, self._nplike)
            nextcarry = ak._v2.index.Index64.empty(head.length * length, self._nplike)

            assert (
                multistarts.nplike is self._nplike
                and multistops.nplike is self._nplike
                and singleoffsets.nplike is self._nplike
                and nextcarry.nplike is self._nplike
                and self._starts.nplike is self._nplike
                and self._stops.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_ListArray_getitem_jagged_expand",
                    multistarts.dtype.type,
                    multistops.dtype.type,
                    singleoffsets.dtype.type,
                    nextcarry.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    multistarts.data,
                    multistops.data,
                    singleoffsets.data,
                    nextcarry.data,
                    self._starts.data,
                    self._stops.data,
                    head.length,
                    length,
                ),
            )
            carried = self._content._carry(nextcarry, True, NestedIndexError)
            down = carried._getitem_next_jagged(
                multistarts, multistops, head._content, tail
            )

            return ak._v2.contents.regulararray.RegularArray(
                down, headlength, 1, None, self._parameters, self._nplike
            )

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def num(self, axis, depth=0):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            out = ak._v2.index.Index64.empty(1, self._nplike)
            out[0] = self.length
            return ak._v2.contents.numpyarray.NumpyArray(out, None, None, self._nplike)[
                0
            ]
        elif posaxis == depth + 1:
            tonum = ak._v2.index.Index64.empty(self.length, self._nplike)
            assert (
                tonum.nplike is self._nplike
                and self._starts.nplike is self._nplike
                and self._stops.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_ListArray_num",
                    tonum.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    tonum.data,
                    self._starts.data,
                    self._stops.data,
                    self.length,
                )
            )
            return ak._v2.contents.numpyarray.NumpyArray(
                tonum, None, None, self._nplike
            )
        else:
            return self.toListOffsetArray64(True).num(posaxis, depth)

    def _offsets_and_flattened(self, axis, depth):
        return self.toListOffsetArray64(True)._offsets_and_flattened(axis, depth)

    def mergeable(self, other, mergebool):
        if not _parameters_equal(self._parameters, other._parameters):
            return False

        if isinstance(
            other,
            (
                ak._v2.contents.emptyarray.EmptyArray,
                ak._v2.contents.unionarray.UnionArray,
            ),
        ):
            return True

        if isinstance(
            other,
            (
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return self.mergeable(other.content, mergebool)

        if isinstance(
            other,
            (
                ak._v2.contents.regulararray.RegularArray,
                ak._v2.contents.listarray.ListArray,
                ak._v2.contents.listoffsetarray.ListOffsetArray,
            ),
        ):
            return self._content.mergeable(other.content, mergebool)

        else:
            return False

    def mergemany(self, others):
        if len(others) == 0:
            return self

        head, tail = self._merging_strategy(others)

        total_length = 0
        for array in head:
            total_length += array.length

        contents = []

        for array in head:
            parameters = ak._v2._util.merge_parameters(
                self._parameters, array._parameters, True
            )

            if isinstance(
                array,
                (
                    ak._v2.contents.listarray.ListArray,
                    ak._v2.contents.listoffsetarray.ListOffsetArray,
                    ak._v2.contents.regulararray.RegularArray,
                ),
            ):
                contents.append(array.content)

            elif isinstance(array, ak._v2.contents.emptyarray.EmptyArray):
                pass
            else:
                raise ValueError(
                    "cannot merge "
                    + type(self).__name__
                    + " with "
                    + type(array).__name__
                    + "."
                )

        tail_contents = contents[1:]
        nextcontent = contents[0].mergemany(tail_contents)

        nextstarts = ak._v2.index.Index64.empty(total_length, self._nplike)
        nextstops = ak._v2.index.Index64.empty(total_length, self._nplike)

        contentlength_so_far = 0
        length_so_far = 0

        for array in head:
            if isinstance(
                array,
                (
                    ak._v2.contents.listarray.ListArray,
                    ak._v2.contents.listoffsetarray.ListOffsetArray,
                ),
            ):
                array_starts = ak._v2.index.Index(array.starts)
                array_stops = ak._v2.index.Index(array.stops)

                assert (
                    nextstarts.nplike is self._nplike
                    and nextstops.nplike is self._nplike
                    and array_starts.nplike is self._nplike
                    and array_stops.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_ListArray_fill",
                        nextstarts.dtype.type,
                        nextstops.dtype.type,
                        array_starts.dtype.type,
                        array_stops.dtype.type,
                    ](
                        nextstarts.data,
                        length_so_far,
                        nextstops.data,
                        length_so_far,
                        array_starts.data,
                        array_stops.data,
                        array.length,
                        contentlength_so_far,
                    )
                )
                contentlength_so_far += array.content.length
                length_so_far += array.length

            elif isinstance(array, ak._v2.contents.regulararray.RegularArray):
                listoffsetarray = array.toListOffsetArray64(True)

                array_starts = ak._v2.index.Index64(listoffsetarray.starts)
                array_stops = ak._v2.index.Index64(listoffsetarray.stops)

                assert (
                    nextstarts.nplike is self._nplike
                    and nextstops.nplike is self._nplike
                    and array_starts.nplike is self._nplike
                    and array_stops.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_ListArray_fill",
                        nextstarts.dtype.type,
                        nextstops.dtype.type,
                        array_starts.dtype.type,
                        array_stops.dtype.type,
                    ](
                        nextstarts.data,
                        length_so_far,
                        nextstops.data,
                        length_so_far,
                        array_starts.data,
                        array_stops.data,
                        listoffsetarray.length,
                        contentlength_so_far,
                    )
                )
                contentlength_so_far += array.content.length
                length_so_far += listoffsetarray.length

            elif isinstance(array, ak._v2.contents.emptyarray.EmptyArray):
                pass

        next = ak._v2.contents.listarray.ListArray(
            nextstarts, nextstops, nextcontent, None, parameters, self._nplike
        )

        if len(tail) == 0:
            return next

        reversed = tail[0]._reverse_merge(next)
        if len(tail) == 1:
            return reversed
        else:
            return reversed.mergemany(tail[1:])

    def fillna(self, value):
        return ListArray(
            self._starts,
            self._stops,
            self._content.fillna(value),
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def _localindex(self, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._localindex_axis0()
        elif posaxis == depth + 1:
            offsets = self._compact_offsets64(True)
            if self._nplike.known_data:
                innerlength = offsets[offsets.length - 1]
            else:
                innerlength = ak._v2._typetracer.UnknownLength
            localindex = ak._v2.index.Index64.empty(innerlength, self._nplike)
            assert localindex.nplike is self._nplike and offsets.nplike is self._nplike
            self._handle_error(
                self._nplike[
                    "awkward_ListArray_localindex",
                    localindex.dtype.type,
                    offsets.dtype.type,
                ](
                    localindex.data,
                    offsets.data,
                    offsets.length - 1,
                )
            )
            return ak._v2.contents.listoffsetarray.ListOffsetArray(
                offsets,
                ak._v2.contents.NumpyArray(localindex),
                self._identifier,
                self._parameters,
                self._nplike,
            )
        else:
            return ak._v2.contents.listarray.ListArray(
                self._starts,
                self._stops,
                self._content._localindex(posaxis, depth + 1),
                self._identifier,
                self._parameters,
                self._nplike,
            )

    def numbers_to_type(self, name):
        return ak._v2.contents.listarray.ListArray(
            self._starts,
            self._stops,
            self._content.numbers_to_type(name),
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def _is_unique(self, negaxis, starts, parents, outlength):
        if self._starts.length == 0:
            return True

        return self.toListOffsetArray64(True)._is_unique(
            negaxis, starts, parents, outlength
        )

    def _unique(self, negaxis, starts, parents, outlength):
        if self._starts.length == 0:
            return self
        return self.toListOffsetArray64(True)._unique(
            negaxis, starts, parents, outlength
        )

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if self._starts.length == 0:
            return ak._v2.contents.NumpyArray(
                self._nplike.empty(0, np.int64), None, None, self._nplike
            )

        next = self.toListOffsetArray64(True)
        out = next._argsort_next(
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )
        return out

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        return self.toListOffsetArray64(True)._sort_next(
            negaxis,
            starts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        return ListOffsetArray._combinations(
            self, n, replacement, recordlookup, parameters, axis, depth
        )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        return self.toListOffsetArray64(True)._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
        )

    def _validityerror(self, path):
        if self._nplike.known_shape and self.stops.length < self.starts.length:
            return f'at {path} ("{type(self)}"): len(stops) < len(starts)'
        assert self.starts.nplike is self._nplike and self.stops.nplike is self._nplike
        error = self._nplike[
            "awkward_ListArray_validity", self.starts.dtype.type, self.stops.dtype.type
        ](
            self.starts.data,
            self.stops.data,
            self.starts.length,
            self._content.length,
        )
        if error.str is not None:
            if error.filename is None:
                filename = ""
            else:
                filename = " (in compiled code: " + error.filename.decode(
                    errors="surrogateescape"
                ).lstrip("\n").lstrip("(")
            message = error.str.decode(errors="surrogateescape")
            return 'at {} ("{}"): {} at i={}{}'.format(
                path, type(self), message, error.id, filename
            )
        else:
            if (
                self.parameter("__array__") == "string"
                or self.parameter("__array__") == "bytestring"
            ):
                return ""
            else:
                return self._content.validityerror(path + ".content")

    def _nbytes_part(self):
        result = (
            self.starts._nbytes_part()
            + self.stops._nbytes_part()
            + self.content._nbytes_part()
        )
        if self.identifier is not None:
            result = result + self.identifier._nbytes_part()
        return result

    def _rpad(self, target, axis, depth, clip):
        if not clip:
            posaxis = self.axis_wrap_if_negative(axis)
            if posaxis == depth:
                return self.rpad_axis0(target, clip)
            elif posaxis == depth + 1:
                min_ = ak._v2.index.Index64.empty(1, self._nplike)
                assert (
                    min_.nplike is self._nplike
                    and self._starts.nplike is self._nplike
                    and self._stops.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_ListArray_min_range",
                        min_.dtype.type,
                        self._starts.dtype.type,
                        self._stops.dtype.type,
                    ](
                        min_.data,
                        self._starts.data,
                        self._stops.data,
                        self._starts.length,
                    )
                )
                # TODO: Replace the kernel call with below code once typtracer supports '-'
                # min_ = self._nplike.min(self._stops.data - self._starts.data)
                if target < min_[0]:
                    return self
                else:
                    tolength = ak._v2.index.Index64.empty(1, self._nplike)
                    assert (
                        tolength.nplike is self._nplike
                        and self._starts.nplike is self._nplike
                        and self._stops.nplike is self._nplike
                    )
                    self._handle_error(
                        self._nplike[
                            "awkward_ListArray_rpad_and_clip_length_axis1",
                            tolength.dtype.type,
                            self._starts.dtype.type,
                            self._stops.dtype.type,
                        ](
                            tolength.data,
                            self._starts.data,
                            self._stops.data,
                            target,
                            self._starts.length,
                        )
                    )

                    index = ak._v2.index.Index64.empty(tolength[0], self._nplike)
                    starts_ = ak._v2.index.Index64.empty(
                        self._starts.length, self._nplike
                    )
                    stops_ = ak._v2.index.Index64.empty(
                        self._stops.length, self._nplike
                    )
                    assert (
                        index.nplike is self._nplike
                        and self._starts.nplike is self._nplike
                        and self._stops.nplike is self._nplike
                        and starts_.nplike is self._nplike
                        and stops_.nplike is self._nplike
                    )
                    self._handle_error(
                        self._nplike[
                            "awkward_ListArray_rpad_axis1",
                            index.dtype.type,
                            self._starts.dtype.type,
                            self._stops.dtype.type,
                            starts_.dtype.type,
                            stops_.dtype.type,
                        ](
                            index.data,
                            self._starts.data,
                            self._stops.data,
                            starts_.data,
                            stops_.data,
                            target,
                            self._starts.length,
                        )
                    )
                    next = ak._v2.contents.indexedoptionarray.IndexedOptionArray(
                        index,
                        self._content,
                        None,
                        None,
                        self._nplike,
                    )

                    return ak._v2.contents.listarray.ListArray(
                        starts_,
                        stops_,
                        next.simplify_optiontype(),
                        None,
                        self._parameters,
                        self._nplike,
                    )
            else:
                return ak._v2.contents.listarray.ListArray(
                    self._starts,
                    self._stops,
                    self._content._rpad(target, posaxis, depth + 1, clip),
                    None,
                    self._parameters,
                    self._nplike,
                )
        else:
            return self.toListOffsetArray64(True)._rpad(target, axis, depth, clip=True)

    def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
        return self.toListOffsetArray64(False)._to_arrow(
            pyarrow, mask_node, validbytes, length, options
        )

    def _to_numpy(self, allow_missing):
        return ak._v2.operations.convert.to_numpy(self.toRegularArray(), allow_missing)

    def _completely_flatten(self, nplike, options):
        if (
            self.parameter("__array__") == "string"
            or self.parameter("__array__") == "bytestring"
        ):
            return [ak._v2.operations.convert.to_numpy(self)]
        else:
            next = self.toListOffsetArray64(False)
            flat = next.content[next.offsets[0] : next.offsets[-1]]
            return flat._completely_flatten(nplike, options)

    def _recursively_apply(
        self, action, depth, depth_context, lateral_context, options
    ):
        if options["return_array"]:

            def continuation():
                return ListArray(
                    self._starts,
                    self._stops,
                    self._content._recursively_apply(
                        action,
                        depth + 1,
                        copy.copy(depth_context),
                        lateral_context,
                        options,
                    ),
                    self._identifier,
                    self._parameters if options["keep_parameters"] else None,
                    self._nplike,
                )

        else:

            def continuation():
                self._content._recursively_apply(
                    action,
                    depth + 1,
                    copy.copy(depth_context),
                    lateral_context,
                    options,
                )

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def packed(self):
        return self.toListOffsetArray64(True).packed()

    def _to_list(self, behavior):
        return ListOffsetArray._to_list(self, behavior)

    def _to_nplike(self, nplike):
        starts = self._starts._to_nplike(nplike)
        stops = self._stops._to_nplike(nplike)
        content = self._content._to_nplike(nplike)
        return ListArray(
            starts,
            stops,
            content,
            identifier=self._identifier,
            parameters=self._parameters,
            nplike=nplike,
        )

    def _to_json(
        self,
        nan_string,
        infinity_string,
        minus_infinity_string,
        complex_real_string,
        complex_imag_string,
    ):
        return ListOffsetArray._to_json(
            self,
            nan_string,
            infinity_string,
            minus_infinity_string,
            complex_real_string,
            complex_imag_string,
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak


def validity_error(array, exception=False):
    """
    Args:
        array (#ak.Array, #ak.Record, #ak.layout.Content, #ak.layout.Record, #ak.ArrayBuilder, #ak.layout.ArrayBuilder):
            Array or record to check.
        exception (bool): If True, validity errors raise exceptions.

    Returns an empty string if there are no errors and a str containing the error message
    if there are.

    Checks for errors in the structure of the array, such as indexes that run
    beyond the length of a node's `content`, etc. Either an error is raised or
    a string describing the error is returned.

    See also #ak.is_valid.
    """
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )
    out = layout.validityerror(path="highlevel")

    if out not in (None, "") and exception:
        raise ValueError(out)
    else:
        return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.forms.emptyform import EmptyForm
from awkward._v2.forms.form import _parameters_equal

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()


class EmptyArray(Content):
    is_NumpyType = True
    is_UnknownType = True

    def __init__(self, identifier=None, parameters=None, nplike=None):
        if nplike is None:
            nplike = numpy
        self._init(identifier, parameters, nplike)

    Form = EmptyForm

    def _form_with_key(self, getkey):
        return self.Form(
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=getkey(self),
        )

    def _to_buffers(self, form, getkey, container, nplike):
        assert isinstance(form, self.Form)

    @property
    def typetracer(self):
        tt = ak._v2._typetracer.TypeTracer.instance()
        return EmptyArray(self._typetracer_identifier(), self._parameters, tt)

    @property
    def length(self):
        return 0

    def _forget_length(self):
        return EmptyArray(self._identifier, self._parameters, self._nplike)

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        extra = self._repr_extra(indent + "    ")
        if len(extra) == 0:
            return indent + pre + "<EmptyArray len='0'/>" + post
        else:
            out = [indent, pre, "<EmptyArray len='0'>"]
            out.extend(extra)
            out.append("\n" + indent + "</EmptyArray>")
            out.append(post)
            return "".join(out)

    def merge_parameters(self, parameters):
        return EmptyArray(
            self._identifier,
            ak._v2._util.merge_parameters(self._parameters, parameters),
            self._nplike,
        )

    def toNumpyArray(self, dtype, nplike=None):
        if nplike is None:
            nplike = self._nplike
        if nplike is None:
            nplike = numpy
        return ak._v2.contents.numpyarray.NumpyArray(
            nplike.empty(0, dtype), self._identifier, self._parameters, nplike
        )

    def __array__(self, **kwargs):
        return numpy.empty((0,))

    def __iter__(self):
        return iter([])

    def _getitem_nothing(self):
        return self

    def _getitem_at(self, where):
        raise NestedIndexError(self, where, "array is empty")

    def _getitem_range(self, where):
        return self

    def _getitem_field(self, where, only_fields=()):
        raise NestedIndexError(self, where, "not an array of records")

    def _getitem_fields(self, where, only_fields=()):
        if len(where) == 0:
            return self._getitem_range(slice(0, 0))
        raise NestedIndexError(self, where, "not an array of records")

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)

        if not carry.nplike.known_shape or carry.length == 0:
            return self
        else:
            if issubclass(exception, NestedIndexError):
                raise exception(self, carry.data, "array is empty")
            else:
                raise exception("array is empty")

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif isinstance(head, int):
            raise NestedIndexError(self, head, "array is empty")

        elif isinstance(head, slice):
            raise NestedIndexError(self, head, "array is empty")

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.index.Index64):
            raise NestedIndexError(self, head, "array is empty")

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            raise NestedIndexError(self, head, "array is empty")

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            raise NestedIndexError(self, head, "array is empty")

        else:
            raise AssertionError(repr(head))

    def num(self, axis, depth=0):
        posaxis = self.axis_wrap_if_negative(axis)

        if posaxis == depth:
            out = ak._v2.index.Index64.empty(1, self._nplike)
            out[0] = self.length
            return ak._v2.contents.numpyarray.NumpyArray(out, None, None, self._nplike)[
                0
            ]
        else:
            out = ak._v2.index.Index64.empty(0, self._nplike)
            return ak._v2.contents.numpyarray.NumpyArray(out, None, None, self._nplike)

    def _offsets_and_flattened(self, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            raise np.AxisError(self, "axis=0 not allowed for flatten")
        else:
            offsets = ak._v2.index.Index64.zeros(1, self._nplike)
            return (offsets, EmptyArray(None, self._parameters, self._nplike))

    def mergeable(self, other, mergebool):
        if not _parameters_equal(self._parameters, other._parameters):
            return False
        return True

    def mergemany(self, others):
        if len(others) == 0:
            return self

        elif len(others) == 1:
            return others[0]

        else:
            tail_others = others[1:]
            return others[0].mergemany(tail_others)

    def fillna(self, value):
        return EmptyArray(None, self._parameters, self._nplike)

    def _localindex(self, axis, depth):
        return ak._v2.contents.numpyarray.NumpyArray(
            self._nplike.empty(0, np.int64), None, None, self._nplike
        )

    def numbers_to_type(self, name):
        return ak._v2.contents.emptyarray.EmptyArray(
            self._identifier, self._parameters, self._nplike
        )

    def _is_unique(self, negaxis, starts, parents, outlength):
        return True

    def _unique(self, negaxis, starts, parents, outlength):
        return self

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        as_numpy = self.toNumpyArray(np.float64)
        return as_numpy._argsort_next(
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        return self

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        return ak._v2.contents.emptyarray.EmptyArray(
            self._identifier, self._parameters, self._nplike
        )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        as_numpy = self.toNumpyArray(reducer.preferred_dtype)
        return as_numpy._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
        )

    def _validityerror(self, path):
        return ""

    def _nbytes_part(self):
        if self.identifier is not None:
            return self.identifier._nbytes_part()
        return 0

    def _rpad(self, target, axis, depth, clip):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis != depth:
            raise np.AxisError(f"axis={axis} exceeds the depth of this array({depth})")
        else:
            return self.rpad_axis0(target, True)

    def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
        if options["emptyarray_to"] is None:
            return pyarrow.Array.from_buffers(
                ak._v2._connect.pyarrow.to_awkwardarrow_type(
                    pyarrow.null(), options["extensionarray"], mask_node, self
                ),
                length,
                [
                    ak._v2._connect.pyarrow.to_validbits(validbytes),
                ],
                null_count=length,
            )

        else:
            dtype = np.dtype(options["emptyarray_to"])
            next = ak._v2.contents.numpyarray.NumpyArray(
                numpy.empty(length, dtype),
                self._identifier,
                self._parameters,
                nplike=self.nplike,
            )
            return next._to_arrow(pyarrow, mask_node, validbytes, length, options)

    def _to_numpy(self, allow_missing):
        return self._nplike.empty(0, dtype=np.float64)

    def _completely_flatten(self, nplike, options):
        return []

    def _recursively_apply(
        self, action, depth, depth_context, lateral_context, options
    ):
        if options["return_array"]:

            def continuation():
                if options["keep_parameters"]:
                    return self
                else:
                    return EmptyArray(self._identifier, None, self._nplike)

        else:

            def continuation():
                pass

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def packed(self):
        return self

    def _to_list(self, behavior):
        return []

    def _to_nplike(self, nplike):
        return EmptyArray(self._identifier, self._parameters, nplike=nplike)

    def _to_json(
        self,
        nan_string,
        infinity_string,
        minus_infinity_string,
        complex_real_string,
        complex_imag_string,
    ):
        return []



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

# v2: no change; keep this file.

import ctypes
import platform

import awkward._kernel_signatures_cuda

if platform.system() == "Windows":
    raise ValueError(
        "There is no support for Awkward CUDA on Windows. Please use a Linux system to use CUDA "
        "accelerated Awkward Arrays."
    )
elif platform.system() == "Darwin":
    raise ValueError(
        "There is no support for Awkward CUDA on MacOS. Please use a Linux system to use CUDA "
        "accelerated Awkward Arrays."
    )
else:
    try:
        import awkward_cuda_kernels
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            """to use CUDA awkward arrays in Python, install the 'awkward[cuda]' package with:
    pip install awkward[cuda] --upgrade
or
    conda install -c conda-forge awkward-cuda-kernels"""
        ) from None

lib = ctypes.cdll.LoadLibrary(awkward_cuda_kernels.shared_library_path)
kernel = awkward._kernel_signatures_cuda.by_signature(lib)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def fields(array):

    """
    Extracts record fields or tuple slot numbers from `array` (many types
    supported, including all Awkward Arrays and Records).

    If the array contains nested records, only the outermost record is
    queried. If it contains tuples instead of records, this function outputs
    string representations of integers, such as `"0"`, `"1"`, `"2"`, etc.
    The records or tuples may be within multiple layers of nested lists.

    If the array contains neither tuples nor records, this returns an empty
    list.
    """
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=True, allow_other=False
    )
    return layout.fields



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def to_backend(array, backend, highlevel=True, behavior=None):
    """
    Args:
        array: Data to convert to a specified `backend` set.
        backend (`"cpu"` or `"cuda"`): If `"cpu"`, the array structure is
            recursively copied (if need be) to main memory for use with
            the default `libawkward-cpu-kernels.so`; if `"cuda"`, the
            structure is copied to the GPU(s) for use with
            `libawkward-cuda-kernels.so`.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Converts an array from `"cpu"`, `"cuda"`, or `"mixed"` kernels to `"cpu"`
    or `"cuda"`.

    An array is `"mixed"` if some components are set to use the `"cpu"` backend and
    others are set to use the `"cuda"` backend. Mixed arrays can't be used in any
    operations, and two arrays set to different backends can't be used in the
    same operation.

    Any components that are already in the desired backend are viewed,
    rather than copied, so this operation can be an inexpensive way to ensure
    that an array is ready for a particular library.

    To use `"cuda"`, the package
    [awkward-cuda-kernels](https://pypi.org/project/awkward-cuda-kernels)
    be installed, either by

        pip install awkward-cuda-kernels

    or as an optional dependency with

        pip install awkward[cuda] --upgrade

    It is only available for Linux as a binary wheel, and only supports Nvidia
    GPUs (it is written in CUDA).

    See #ak.kernels.
    """
    layout = ak._v2.operations.convert.to_layout(
        array,
        allow_record=True,
        allow_other=True,
    )
    backend_layout = layout.to_backend(backend)
    if highlevel:
        return ak._v2.Array(backend_layout, behavior=behavior)
    return backend_layout



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def run_lengths(array, highlevel=True, behavior=None):
    raise NotImplementedError


#     """
#     Args:
#         array: Data containing runs of numbers to count.
#         highlevel (bool): If True, return an #ak.Array; otherwise, return
#             a low-level #ak.layout.Content subclass.
#         behavior (None or dict): Custom #ak.behavior for the output array, if
#             high-level.

#     Computes the lengths of sequences of identical values at the deepest level
#     of nesting, returning an array with the same structure but with `int64` type.

#     For example,

#         >>> array = ak.Array([1.1, 1.1, 1.1, 2.2, 3.3, 3.3, 4.4, 4.4, 5.5])
#         >>> ak.run_lengths(array)
#         <Array [3, 1, 2, 2, 1] type='5 * int64'>

#     There are 3 instances of 1.1, followed by 1 instance of 2.2, 2 instances of 3.3,
#     2 instances of 4.4, and 1 instance of 5.5.

#     The order and uniqueness of the input data doesn't matter,

#         >>> array = ak.Array([1.1, 1.1, 1.1, 5.5, 4.4, 4.4, 1.1, 1.1, 5.5])
#         >>> ak.run_lengths(array)
#         <Array [3, 1, 2, 2, 1] type='5 * int64'>

#     just the difference between each value and its neighbors.

#     The data can be nested, but runs don't cross list boundaries.

#         >>> array = ak.Array([[1.1, 1.1, 1.1, 2.2, 3.3], [3.3, 4.4], [4.4, 5.5]])
#         >>> ak.run_lengths(array)
#         <Array [[3, 1, 1], [1, 1], [1, 1]] type='3 * var * int64'>

#     This function recognizes strings as distinguishable values.

#         >>> array = ak.Array([["one", "one"], ["one", "two", "two"], ["three", "two", "two"]])
#         >>> ak.run_lengths(array)
#         <Array [[2], [1, 2], [1, 2]] type='3 * var * int64'>

#     Note that this can be combined with #ak.argsort and #ak.unflatten to compute
#     a "group by" operation:

#         >>> array = ak.Array([{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 1, "y": 1.1},
#         ...                   {"x": 3, "y": 3.3}, {"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}])
#         >>> sorted = array[ak.argsort(array.x)]
#         >>> sorted.x
#         <Array [1, 1, 1, 2, 2, 3] type='6 * int64'>
#         >>> ak.run_lengths(sorted.x)
#         <Array [3, 2, 1] type='3 * int64'>
#         >>> ak.unflatten(sorted, ak.run_lengths(sorted.x)).tolist()
#         [[{'x': 1, 'y': 1.1}, {'x': 1, 'y': 1.1}, {'x': 1, 'y': 1.1}],
#          [{'x': 2, 'y': 2.2}, {'x': 2, 'y': 2.2}],
#          [{'x': 3, 'y': 3.3}]]

#     Unlike a database "group by," this operation can be applied in bulk to many sublists
#     (though the run lengths need to be fully flattened to be used as `counts` for
#     #ak.unflatten, and you need to specify `axis=-1` as the depth).

#         >>> array = ak.Array([[{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 1, "y": 1.1}],
#         ...                   [{"x": 3, "y": 3.3}, {"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}]])
#         >>> sorted = array[ak.argsort(array.x)]
#         >>> sorted.x
#         <Array [[1, 1, 2], [1, 2, 3]] type='2 * var * int64'>
#         >>> ak.run_lengths(sorted.x)
#         <Array [[2, 1], [1, 1, 1]] type='2 * var * int64'>
#         >>> counts = ak.flatten(ak.run_lengths(sorted.x), axis=None)
#         >>> ak.unflatten(sorted, counts, axis=-1).tolist()
#         [[[{'x': 1, 'y': 1.1}, {'x': 1, 'y': 1.1}],
#           [{'x': 2, 'y': 2.2}]],
#          [[{'x': 1, 'y': 1.1}],
#           [{'x': 2, 'y': 2.2}],
#           [{'x': 3, 'y': 3.3}]]]

#     See also #ak.num, #ak.argsort, #ak.unflatten.
#     """
#     nplike = ak.nplike.of(array)

#     def lengths_of(data, offsets):
#         if len(data) == 0:
#             return nplike.empty(0, np.int64), offsets
#         else:
#             diffs = data[1:] != data[:-1]
#             if isinstance(diffs, ak._v2.highlevel.Array):
#                 diffs = nplike.asarray(diffs)
#             if offsets is not None:
#                 diffs[offsets[1:-1] - 1] = True
#             positions = nplike.nonzero(diffs)[0]
#             full_positions = nplike.empty(len(positions) + 2, np.int64)
#             full_positions[0] = 0
#             full_positions[-1] = len(data)
#             full_positions[1:-1] = positions + 1
#             nextcontent = full_positions[1:] - full_positions[:-1]
#             if offsets is None:
#                 nextoffsets = None
#             else:
#                 nextoffsets = nplike.searchsorted(full_positions, offsets, side="left")
#             return nextcontent, nextoffsets

#     def getfunction(layout):
#         if layout.branch_depth == (False, 1):
#             if isinstance(layout, ak._v2._util.indexedtypes):
#                 layout = layout.project()

#             if (
#                 layout.parameter("__array__") == "string"
#                 or layout.parameter("__array__") == "bytestring"
#             ):
#                 nextcontent, _ = lengths_of(ak._v2.highlevel.Array(layout), None)
#                 return lambda: ak._v2.contents.NumpyArray(nextcontent)

#             if not isinstance(layout, (ak._v2.contents.NumpyArray, ak._v2.contents.EmptyArray)):
#                 raise NotImplementedError(
#                     "run_lengths on "
#                     + type(layout).__name__
#
#                 )

#             nextcontent, _ = lengths_of(nplike.asarray(layout), None)
#             return lambda: ak._v2.contents.NumpyArray(nextcontent)

#         elif layout.branch_depth == (False, 2):
#             if isinstance(layout, ak._v2._util.indexedtypes):
#                 layout = layout.project()

#             if not isinstance(layout, ak._v2._util.listtypes):
#                 raise NotImplementedError(
#                     "run_lengths on "
#                     + type(layout).__name__
#
#                 )

#             if (
#                 layout.content.parameter("__array__") == "string"
#                 or layout.content.parameter("__array__") == "bytestring"
#             ):
#                 listoffsetarray = layout.toListOffsetArray64(False)
#                 offsets = nplike.asarray(listoffsetarray.offsets)
#                 content = listoffsetarray.content[offsets[0] : offsets[-1]]

#                 if isinstance(content, ak._v2._util.indexedtypes):
#                     content = content.project()

#                 nextcontent, nextoffsets = lengths_of(
#                     ak._v2.highlevel.Array(content), offsets - offsets[0]
#                 )
#                 return lambda: ak._v2.contents.ListOffsetArray64(
#                     ak._v2.index.Index64(nextoffsets), ak._v2.contents.NumpyArray(nextcontent)
#                 )

#             listoffsetarray = layout.toListOffsetArray64(False)
#             offsets = nplike.asarray(listoffsetarray.offsets)
#             content = listoffsetarray.content[offsets[0] : offsets[-1]]

#             if isinstance(content, ak._v2._util.indexedtypes):
#                 content = content.project()

#             if not isinstance(content, (ak._v2.contents.NumpyArray, ak._v2.contents.EmptyArray)):
#                 raise NotImplementedError(
#                     "run_lengths on "
#                     + type(layout).__name__
#                     + " with content "
#                     + type(content).__name__
#
#                 )

#             nextcontent, nextoffsets = lengths_of(
#                 nplike.asarray(content), offsets - offsets[0]
#             )
#             return lambda: ak._v2.contents.ListOffsetArray64(
#                 ak._v2.index.Index64(nextoffsets), ak._v2.contents.NumpyArray(nextcontent)
#             )

#         else:
#             return None

#     layout = ak._v2.operations.convert.to_layout(
#         array, allow_record=False, allow_other=False
#     )

#     if isinstance(layout, ak.partition.PartitionedArray):   # NO PARTITIONED ARRAY
#         if len(layout.partitions) != 0 and layout.partitions[0].branch_depth == (
#             False,
#             1,
#         ):
#             out = ak._v2._util.recursively_apply(
#                 layout.toContent(),
#                 getfunction,
#                 pass_depth=False,
#                 pass_user=False,
#             )
#         else:
#             outparts = []
#             for part in layout.partitions:
#                 outparts.append(
#                     ak._v2._util.recursively_apply(
#                         part,
#                         getfunction,
#                         pass_depth=False,
#                         pass_user=False,
#                     )
#                 )
#             out = ak.partition.IrregularlyPartitionedArray(outparts)   # NO PARTITIONED ARRAY
#     else:
#         out = ak._v2._util.recursively_apply(
#             layout,
#             getfunction,
#             pass_depth=False,
#             pass_user=False,
#         )

#     return ak._v2._util.maybe_wrap_like(out, array, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

# from __future__ import absolute_import

# import awkward as ak

# np = ak.nplike.NumpyMetadata.instance()


# class CategoricalBehavior(ak._v2.highlevel.Array):
#     __name__ = "Array"


# ak._v2.behavior["categorical"] = CategoricalBehavior


# class _HashableDict(object):
#     def __init__(self, obj):
#         self.keys = tuple(sorted(obj))
#         self.values = tuple(_hashable(obj[k]) for k in self.keys)
#         self.hash = hash((_HashableDict,) + self.keys, self.values)

#     def __hash__(self):
#         return self.hash

#     def __eq__(self, other):
#         return (
#             isinstance(other, _HashableDict)
#             and self.keys == other.keys
#             and self.values == other.values
#         )


# class _HashableList(object):
#     def __init__(self, obj):
#         self.values = tuple(obj)
#         self.hash = hash((_HashableList,) + self.values)

#     def __hash__(self):
#         return self.hash

#     def __eq__(self, other):
#         return isinstance(other, _HashableList) and self.values == other.values


# def _hashable(obj):
#     if isinstance(obj, dict):
#         return _HashableDict(obj)
#     elif isinstance(obj, tuple):
#         return tuple(_hashable(x) for x in obj)
#     elif isinstance(obj, list):
#         return _HashableList(obj)
#     else:
#         return obj


# def _categorical_equal(one, two):
#     behavior = ak._v2._util.behaviorof(one, two)

#     one, two = one.layout, two.layout

#     assert isinstance(one, ak._v2._util.indexedtypes + ak._v2._util.indexedoptiontypes)
#     assert isinstance(two, ak._v2._util.indexedtypes + ak._v2._util.indexedoptiontypes)
#     assert one.parameter("__array__") == "categorical"
#     assert two.parameter("__array__") == "categorical"

#     one_index = ak.nplike.numpy.asarray(one.index)
#     two_index = ak.nplike.numpy.asarray(two.index)
#     one_content = ak._v2._util.wrap(one.content, behavior)
#     two_content = ak._v2._util.wrap(two.content, behavior)

#     if len(one_content) == len(two_content) and ak._v2.operations.reducers.all(
#         one_content == two_content, axis=None
#     ):
#         one_mapped = one_index

#     else:
#         one_list = ak._v2.operations.convert.to_list(one_content)
#         two_list = ak._v2.operations.convert.to_list(two_content)
#         one_hashable = [_hashable(x) for x in one_list]
#         two_hashable = [_hashable(x) for x in two_list]
#         two_lookup = {x: i for i, x in enumerate(two_hashable)}

#         one_to_two = ak.nplike.numpy.empty(len(one_hashable) + 1, dtype=np.int64)
#         for i, x in enumerate(one_hashable):
#             one_to_two[i] = two_lookup.get(x, len(two_hashable))
#         one_to_two[-1] = -1

#         one_mapped = one_to_two[one_index]

#     out = one_mapped == two_index
#     return ak._v2._util.wrap(ak._v2.contents.NumpyArray(out), ak._v2._util.behaviorof(one, two))


# ak._v2.behavior[ak.nplike.numpy.equal, "categorical", "categorical"] = _categorical_equal


# def _apply_ufunc(ufunc, method, inputs, kwargs):
#     nextinputs = []
#     for x in inputs:
#         if isinstance(x, ak._v2.highlevel.Array) and isinstance(
#             x.layout, ak._v2._util.indexedtypes
#         ):
#             nextinputs.append(
#                 ak._v2.highlevel.Array(x.layout.project(), behavior=ak._v2._util.behaviorof(x))
#             )
#         else:
#             nextinputs.append(x)

#     return getattr(ufunc, method)(*nextinputs, **kwargs)


# ak._v2.behavior[ak.nplike.numpy.ufunc, "categorical"] = _apply_ufunc


# def is_categorical(array):
#     """
#     Args:
#         array: A possibly-categorical Awkward Array.

#     If the `array` is categorical (contains #ak.layout.IndexedArray or
#     #ak.layout.IndexedOptionArray labeled with parameter
#     `"__array__" = "categorical"`), then this function returns True;
#     otherwise, it returns False.

#     See also #ak.categories, #ak.to_categorical, #ak.from_categorical.
#     """

#     layout = ak._v2.operations.convert.to_layout(
#         array, allow_record=False, allow_other=False
#     )
#     return layout.purelist_parameter("__array__") == "categorical"


# def categories(array, highlevel=True):
#     """
#     Args:
#         array: A possibly-categorical Awkward Array.
#         highlevel (bool): If True, return an #ak.Array; otherwise, return
#             a low-level #ak.layout.Content subclass.

#     If the `array` is categorical (contains #ak.layout.IndexedArray or
#     #ak.layout.IndexedOptionArray labeled with parameter
#     `"__array__" = "categorical"`), then this function returns its categories.

#     See also #ak.is_categorical, #ak.to_categorical, #ak.from_categorical.
#     """

#     output = [None]

#     def getfunction(layout):
#         if layout.parameter("__array__") == "categorical":
#             output[0] = layout.content
#             return lambda: layout

#         else:
#             return None

#     layout = ak._v2.operations.convert.to_layout(
#         array, allow_record=False, allow_other=False
#     )
#     ak._v2._util.recursively_apply(layout, getfunction, pass_depth=False)

#     if output[0] is None:
#         return None
#     elif highlevel:
#         return ak._v2._util.wrap(output[0], ak._v2._util.behaviorof(array))
#     else:
#         return output[0]


# def to_categorical(array, highlevel=True):
#     """
#     Args:
#         array: Data convertible to an Awkward Array
#         highlevel (bool): If True, return an #ak.Array; otherwise, return
#             a low-level #ak.layout.Content subclass.

#     Creates a categorical dataset, which has the following properties:

#        * only distinct values (categories) are stored in their entirety,
#        * pointers to those distinct values are represented by integers
#          (an #ak.layout.IndexedArray or #ak.layout.IndexedOptionArray
#          labeled with parameter `"__array__" = "categorical"`.

#     This is equivalent to R's "factor", Pandas's "categorical", and
#     Arrow/Parquet's "dictionary encoding." It differs from generic uses of
#     #ak.layout.IndexedArray and #ak.layout.IndexedOptionArray in Awkward
#     Arrays by the guarantee of no duplicate categories and the `"categorical"`
#     parameter.

#         >>> array = ak.Array([["one", "two", "three"], [], ["three", "two"]])
#         >>> categorical = ak.to_categorical(array)
#         >>> categorical
#         <Array [['one', 'two', ... 'three', 'two']] type='3 * var * categorical[type=str...'>
#         >>> ak.type(categorical)
#         3 * var * categorical[type=string]
#         >>> ak.to_list(categorical) == ak.to_list(array)
#         True
#         >>> ak.categories(categorical)
#         <Array ['one', 'two', 'three'] type='3 * string'>
#         >>> ak.is_categorical(categorical)
#         True
#         >>> ak.from_categorical(categorical)
#         <Array [['one', 'two', ... 'three', 'two']] type='3 * var * string'>

#     This function descends through nested lists, but not into the fields of
#     records, so records can be categories. To make categorical record
#     fields, split up the record, apply this function to each desired field,
#     and #ak.zip the results together.

#         >>> records = ak.Array([
#         ...     {"x": 1.1, "y": "one"},
#         ...     {"x": 2.2, "y": "two"},
#         ...     {"x": 3.3, "y": "three"},
#         ...     {"x": 2.2, "y": "two"},
#         ...     {"x": 1.1, "y": "one"}
#         ... ])
#         >>> records
#         <Array [{x: 1.1, y: 'one'}, ... y: 'one'}] type='5 * {"x": float64, "y": string}'>
#         >>> categorical_records = ak.zip({
#         ...     "x": ak.to_categorical(records["x"]),
#         ...     "y": ak.to_categorical(records["y"]),
#         ... })
#         >>> categorical_records
#         <Array [{x: 1.1, y: 'one'}, ... y: 'one'}] type='5 * {"x": categorical[type=floa...'>
#         >>> ak.type(categorical_records)
#         5 * {"x": categorical[type=float64], "y": categorical[type=string]}
#         >>> ak.to_list(categorical_records) == ak.to_list(records)
#         True

#     The check for uniqueness is currently implemented in a Python loop, so
#     conversion to categorical should be regarded as expensive. (This can
#     change, but it would always be an _n log(n)_ operation.)

#     See also #ak.is_categorical, #ak.categories, #ak.from_categorical.
#     """

#     def getfunction(layout):
#         if layout.purelist_depth == 1:
#             if isinstance(layout, ak._v2._util.optiontypes):
#                 layout = layout.simplify()

#             if isinstance(layout, ak._v2._util.indexedoptiontypes):
#                 content = layout.content
#                 cls = ak._v2.contents.IndexedOptionArray64
#             elif isinstance(layout, ak._v2._util.indexedtypes):
#                 content = layout.content
#                 cls = ak._v2.contents.IndexedArray64
#             elif isinstance(layout, ak._v2._util.optiontypes):
#                 content = layout.content
#                 cls = ak._v2.contents.IndexedOptionArray64
#             else:
#                 content = layout
#                 cls = ak._v2.contents.IndexedArray64

#             content_list = ak._v2.operations.convert.to_list(content)
#             hashable = [_hashable(x) for x in content_list]

#             lookup = {}
#             is_first = ak.nplike.numpy.empty(len(hashable), dtype=np.bool_)
#             mapping = ak.nplike.numpy.empty(len(hashable), dtype=np.int64)
#             for i, x in enumerate(hashable):
#                 if x in lookup:
#                     is_first[i] = False
#                     mapping[i] = lookup[x]
#                 else:
#                     is_first[i] = True
#                     lookup[x] = j = len(lookup)
#                     mapping[i] = j

#             if isinstance(layout, ak._v2._util.indexedoptiontypes):
#                 original_index = ak.nplike.numpy.asarray(layout.index)
#                 index = mapping[original_index]
#                 index[original_index < 0] = -1
#                 index = ak._v2.index.Index64(index)

#             elif isinstance(layout, ak._v2._util.indexedtypes):
#                 original_index = ak.nplike.numpy.asarray(layout.index)
#                 index = ak._v2.index.Index64(mapping[original_index])

#             elif isinstance(layout, ak._v2._util.optiontypes):
#                 mask = ak.nplike.numpy.asarray(layout.bytemask())
#                 mapping[mask.view(np.bool_)] = -1
#                 index = ak._v2.index.Index64(mapping)

#             else:
#                 index = ak._v2.index.Index64(mapping)

#             out = cls(index, content[is_first], parameters={"__array__": "categorical"})
#             return lambda: out

#         else:
#             return None

#     layout = ak._v2.operations.convert.to_layout(
#         array, allow_record=False, allow_other=False
#     )
#     out = ak._v2._util.recursively_apply(layout, getfunction, pass_depth=False)
#     if highlevel:
#         return ak._v2._util.wrap(out, ak._v2._util.behaviorof(array))
#     else:
#         return out


# def from_categorical(array, highlevel=True):
#     """
#     Args:
#         array: Awkward Array from which to remove the 'categorical' parameter.
#         highlevel (bool): If True, return an #ak.Array; otherwise, return
#             a low-level #ak.layout.Content subclass.

#     This function replaces categorical data with non-categorical data (by
#     removing the label that declares it as such).

#     This is a metadata-only operation; the running time does not scale with the
#     size of the dataset. (Conversion to categorical is expensive; conversion
#     from categorical is cheap.)

#     See also #ak.is_categorical, #ak.categories, #ak.to_categorical,
#     #ak.from_categorical.
#     """

#     def getfunction(layout):
#         if layout.parameter("__array__") == "categorical":
#             out = ak._v2.operations.structure.with_parameter(
#                 layout, "__array__", None, highlevel=False
#             )
#             return lambda: out

#         else:
#             return None

#     layout = ak._v2.operations.convert.to_layout(
#         array, allow_record=False, allow_other=False
#     )
#     out = ak._v2._util.recursively_apply(layout, getfunction, pass_depth=False)
#     if highlevel:
#         return ak._v2._util.wrap(out, ak._v2._util.behaviorof(array))
#     else:
#         return out


# __all__ = [
#     x for x in list(globals()) if not x.startswith("_") and x not in ("ak", "np")
# ]


# def __dir__():
#     return __all__
